"""The painter module provides different painters for parts of the canvas.

Painters can be swapped in and out.

Each painter takes care of a layer in the canvas (such as grid, items
and handles).
"""

from typing import Collection, Dict, Optional

from typing_extensions import Protocol

from gaphas.geometry import Rectangle
from gaphas.item import Item
from gaphas.types import CairoContext


class Painter(Protocol):
    """Painter interface."""

    def paint(
        self, items: Collection[Item], cairo: CairoContext
    ) -> Optional[Dict[Item, Rectangle]]:
        """Do the paint action (called from the View)."""
        pass


class ItemPainterType(Protocol):
    def paint_item(self, item: Item, cairo: CairoContext) -> None:
        """Draw a single item."""

    def paint(self, items: Collection[Item], cairo: CairoContext) -> None:
        """Do the paint action (called from the View)."""
        pass



"""Geometry functions.

Rectangle is a utility class for working with rectangles (unions and
intersections).

A point is represented as a tuple `(x, y)`.
"""
from __future__ import annotations

from math import sqrt
from typing import Iterable, Optional, Tuple, Union

Point = Tuple[float, float]  # x, y
Rect = Tuple[float, float, float, float]  # x, y, width, height


class Rectangle:
    """Python Rectangle implementation. Rectangles can be added (union),
    substituted (intersection) and points and rectangles can be tested to be in
    the rectangle.

    >>> r1= Rectangle(1,1,5,5)
    >>> r2 = Rectangle(3,3,6,7)

    Test if two rectangles intersect:

    >>> if r1 - r2: 'yes'
    'yes'

    >>> r1, r2 = Rectangle(1,2,3,4), Rectangle(1,2,3,4)
    >>> r1 == r2
    True

    >>> r = Rectangle(-5, 3, 10, 8)
    >>> r.width = 2
    >>> r
    Rectangle(-5, 3, 2, 8)

    >>> r = Rectangle(-5, 3, 10, 8)
    >>> r.height = 2
    >>> r
    Rectangle(-5, 3, 10, 2)
    """

    def __init__(
        self,
        x: float = 0,
        y: float = 0,
        width: Optional[float] = None,
        height: Optional[float] = None,
        x1: float = 0,
        y1: float = 0,
    ):
        if width is None:
            self.x = min(x, x1)
            self.width = abs(x1 - x)
        else:
            self.x = x
            self.width = width
        if height is None:
            self.y = min(y, y1)
            self.height = abs(y1 - y)
        else:
            self.y = y
            self.height = height

    @property
    def x1(self) -> float:
        return self.x + self.width

    @x1.setter
    def x1(self, x1: float) -> None:
        width = x1 - self.x
        width = max(width, 0)
        self.width = width

    @property
    def y1(self) -> float:
        return self.y + self.height

    @y1.setter
    def y1(self, y1: float) -> None:
        height = y1 - self.y
        height = max(height, 0)
        self.height = height

    def expand(self, delta: float) -> None:
        """
        >>> r = Rectangle(-5, 3, 10, 8)
        >>> r.expand(5)
        >>> r
        Rectangle(-10, -2, 20, 18)
        """
        self.x -= delta
        self.y -= delta
        self.width += delta * 2
        self.height += delta * 2

    def tuple(self) -> Tuple[float, float, float, float]:
        """A type safe version of `tuple(rectangle)`."""
        return (self.x, self.y, self.width, self.height)

    def __repr__(self) -> str:
        """
        >>> Rectangle(5,7,20,25)
        Rectangle(5, 7, 20, 25)
        """
        if self:
            return f"{self.__class__.__name__}({self.x}, {self.y}, {self.width}, {self.height})"
        return f"{self.__class__.__name__}()"

    def __iter__(self) -> Iterable[float]:
        """
        >>> tuple(Rectangle(1,2,3,4))
        (1, 2, 3, 4)
        """
        return iter((self.x, self.y, self.width, self.height))

    def __getitem__(self, index: int) -> float:
        """
        >>> Rectangle(1,2,3,4)[1]
        2
        """
        return (self.x, self.y, self.width, self.height)[index]

    def __bool__(self) -> bool:
        """
        >>> r=Rectangle(1,2,3,4)
        >>> if r: 'yes'
        'yes'
        >>> r = Rectangle(1,1,0,0)
        >>> if r: 'no'
        """
        return self.width > 0 and self.height > 0

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Rectangle)
            and self.x == other.x
            and self.y == other.y
            and self.width == other.width
            and self.height == self.height
        )

    def __add__(self, obj: Union[Rectangle, Rect]) -> Rectangle:
        """Create a new Rectangle is the union of the current rectangle with
        another Rectangle, tuple `(x,y)` or tuple `(x, y, width, height)`.

        >>> r=Rectangle(5, 7, 20, 25)
        >>> r + (0, 0)
        Traceback (most recent call last):
          ...
        TypeError: Can only add Rectangle or tuple (x, y, width, height), not (0, 0).
        >>> r + (20, 30, 40, 50)
        Rectangle(5, 7, 55, 73)
        """
        return Rectangle(self.x, self.y, self.width, self.height).__iadd__(obj)

    def __iadd__(self, obj: Union[Rectangle, Rect]) -> Rectangle:
        """
        >>> r = Rectangle()
        >>> r += Rectangle(5, 7, 20, 25)
        >>> r += (0, 0, 30, 10)
        >>> r
        Rectangle(0, 0, 30, 32)
        >>> r += 'aap'
        Traceback (most recent call last):
          ...
        TypeError: Can only add Rectangle or tuple (x, y, width, height), not 'aap'.
        """
        try:
            x, y, width, height = obj  # type: ignore[misc]
        except ValueError:
            raise TypeError(
                f"Can only add Rectangle or tuple (x, y, width, height), not {repr(obj)}."
            )
        x1, y1 = x + width, y + height
        if self:
            ox1, oy1 = self.x + self.width, self.y + self.height
            self.x = min(self.x, x)
            self.y = min(self.y, y)
            self.x1 = max(ox1, x1)
            self.y1 = max(oy1, y1)
        else:
            self.x, self.y, self.width, self.height = x, y, width, height
        return self

    def __sub__(self, obj: Union[Rectangle, Rect]) -> Rectangle:
        """Create a new Rectangle is the union of the current rectangle with
        another Rectangle or tuple (x, y, width, height).

        >>> r = Rectangle(5, 7, 20, 25)
        >>> r - (20, 30, 40, 50)
        Rectangle(20, 30, 5, 2)
        >>> r - (30, 40, 40, 50)
        Rectangle()
        """
        return Rectangle(self.x, self.y, self.width, self.height).__isub__(obj)

    def __isub__(self, obj: Union[Rectangle, Rect]) -> Rectangle:
        """
        >>> r = Rectangle()
        >>> r -= Rectangle(5, 7, 20, 25)
        >>> r -= (0, 0, 30, 10)
        >>> r
        Rectangle(5, 7, 20, 3)
        >>> r -= 'aap'
        Traceback (most recent call last):
          ...
        TypeError: Can only subtract Rectangle or tuple (x, y, width, height), not 'aap'.
        """
        try:
            x, y, width, height = obj  # type: ignore[misc]
        except ValueError:
            raise TypeError(
                f"Can only subtract Rectangle or tuple (x, y, width, height), not {repr(obj)}."
            )
        x1, y1 = x + width, y + height

        if self:
            ox1, oy1 = self.x + self.width, self.y + self.height
            self.x = max(self.x, x)
            self.y = max(self.y, y)
            self.x1 = min(ox1, x1)
            self.y1 = min(oy1, y1)
        else:
            self.x, self.y, self.width, self.height = x, y, width, height

        return self

    def __contains__(self, obj: Union[Rectangle, Rect, Point]) -> bool:
        """Check if a point `(x, y)` in inside rectangle `(x, y, width,
        height)` or if a rectangle instance is inside with the rectangle.

        >>> r=Rectangle(10, 5, 12, 12)
        >>> (0, 0) in r
        False
        >>> (10, 6) in r
        True
        >>> (12, 12) in r
        True
        >>> (100, 4) in r
        False
        >>> (11, 6, 5, 5) in r
        True
        >>> (11, 6, 15, 15) in r
        False
        >>> Rectangle(11, 6, 5, 5) in r
        True
        >>> Rectangle(11, 6, 15, 15) in r
        False
        >>> 'aap' in r
        Traceback (most recent call last):
          ...
        TypeError: Should compare to Rectangle, tuple (x, y, width, height) or point (x, y), not 'aap'.
        """
        try:
            x, y, width, height = obj  # type: ignore[misc]
            x1, y1 = x + width, y + width
        except ValueError:
            # point
            try:
                (x, y) = (x1, y1) = obj  # type: ignore[misc]
            except ValueError:
                raise TypeError(
                    f"Should compare to Rectangle, tuple (x, y, width, height) or point (x, y), not {repr(obj)}."
                )
        return x >= self.x and x1 <= self.x1 and y >= self.y and y1 <= self.y1


def distance_point_point(point1: Point, point2: Point = (0.0, 0.0)) -> float:
    """Return the distance from point ``point1`` to ``point2``.

    >>> f"{distance_point_point((0,0), (1,1)):.3f}"
    '1.414'
    """
    dx = point1[0] - point2[0]
    dy = point1[1] - point2[1]
    return sqrt(dx * dx + dy * dy)


def distance_point_point_fast(point1: Point, point2: Point = (0.0, 0.0)) -> float:
    """Return the distance from point ``point1`` to ``point2``. This version is
    faster than ``distance_point_point()``, but less precise.

    >>> distance_point_point_fast((0,0), (1,1))
    2
    """
    dx = point1[0] - point2[0]
    dy = point1[1] - point2[1]
    return abs(dx) + abs(dy)


def distance_rectangle_point(rect: Rect, point: Point) -> float:
    """Return the distance (fast) from a rectangle ``(x, y, width,height)`` to
    a ``point``.

    >>> distance_rectangle_point(Rectangle(0, 0, 10, 10), (11, -1))
    2
    >>> distance_rectangle_point((0, 0, 10, 10), (11, -1))
    2
    >>> distance_rectangle_point((0, 0, 10, 10), (-1, 11))
    2
    """
    dx = dy = 0.0
    px, py = point
    rx, ry, rw, rh = rect

    if px < rx:
        dx = rx - px
    elif px > rx + rw:
        dx = px - (rx + rw)

    if py < ry:
        dy = ry - py
    elif py > ry + rh:
        dy = py - (ry + rh)

    return abs(dx) + abs(dy)


def point_on_rectangle(rect: Rect, point: Point, border: bool = False) -> Point:
    """
    Return the point on which ``point`` can be projecten on the
    rectangle.  ``border = True`` will make sure the point is bound to
    the border of the reactangle. Otherwise, if the point is in the
    rectangle, it's okay.

    >>> point_on_rectangle(Rectangle(0, 0, 10, 10), (11, -1))
    (10, 0)
    >>> point_on_rectangle((0, 0, 10, 10), (5, 12))
    (5, 10)
    >>> point_on_rectangle(Rectangle(0, 0, 10, 10), (12, 5))
    (10, 5)
    >>> point_on_rectangle(Rectangle(1, 1, 10, 10), (3, 4))
    (3, 4)
    >>> point_on_rectangle(Rectangle(1, 1, 10, 10), (0, 3))
    (1, 3)
    >>> point_on_rectangle(Rectangle(1, 1, 10, 10), (4, 3))
    (4, 3)
    >>> point_on_rectangle(Rectangle(1, 1, 10, 10), (4, 9), border=True)
    (4, 11)
    >>> point_on_rectangle((1, 1, 10, 10), (4, 6), border=True)
    (1, 6)
    >>> point_on_rectangle(Rectangle(1, 1, 10, 10), (5, 3), border=True)
    (5, 1)
    >>> point_on_rectangle(Rectangle(1, 1, 10, 10), (8, 4), border=True)
    (11, 4)
    >>> point_on_rectangle((1, 1, 10, 100), (5, 8), border=True)
    (1, 8)
    >>> point_on_rectangle((1, 1, 10, 100), (5, 98), border=True)
    (5, 101)
    """
    px, py = point
    rx, ry, rw, rh = tuple(rect)
    x_inside = y_inside = False

    if px < rx:
        px = rx
    elif px > rx + rw:
        px = rx + rw
    elif border:
        x_inside = True

    if py < ry:
        py = ry
    elif py > ry + rh:
        py = ry + rh
    elif border:
        y_inside = True

    if x_inside and y_inside:
        # Find point on side closest to the point
        if min(abs(rx - px), abs(rx + rw - px)) > min(abs(ry - py), abs(ry + rh - py)):
            if py < ry + rh / 2.0:
                py = ry
            else:
                py = ry + rh
        else:
            if px < rx + rw / 2.0:
                px = rx
            else:
                px = rx + rw

    return px, py


def distance_line_point(
    line_start: Point, line_end: Point, point: Point
) -> Tuple[float, Point]:
    """Calculate the distance of a ``point`` from a line. The line is marked by
    begin and end point ``line_start`` and ``line_end``.

    A tuple is returned containing the distance and point on the line.

    >>> distance_line_point((0., 0.), (2., 4.), point=(3., 4.))
    (1.0, (2.0, 4.0))
    >>> distance_line_point((0., 0.), (2., 4.), point=(-1., 0.))
    (1.0, (0.0, 0.0))
    >>> distance_line_point((0., 0.), (2., 4.), point=(1., 2.))
    (0.0, (1.0, 2.0))
    >>> d, p = distance_line_point((0., 0.), (2., 4.), point=(2., 2.))
    >>> f"{d:.3f}"
    '0.894'
    >>> f"({p[0]:.3f}, {p[1]:.3f})"
    '(1.200, 2.400)'
    """
    # The original end point:
    true_line_end = line_end

    # "Move" the line, so it "starts" on (0, 0)
    line_end = line_end[0] - line_start[0], line_end[1] - line_start[1]
    point = point[0] - line_start[0], point[1] - line_start[1]

    line_len_sqr = line_end[0] * line_end[0] + line_end[1] * line_end[1]

    # Both points are very near each other.
    if line_len_sqr < 0.0001:
        return distance_point_point(point), line_start

    projlen = (line_end[0] * point[0] + line_end[1] * point[1]) / line_len_sqr

    if projlen < 0.0:
        # Closest point is the start of the line.
        return distance_point_point(point), line_start
    elif projlen > 1.0:
        # Point has a projection after the line_end.
        return distance_point_point(point, line_end), true_line_end
    else:
        # Projection is on the line. multiply the line_end with the projlen
        # factor to obtain the point on the line.
        proj = line_end[0] * projlen, line_end[1] * projlen
        return (
            distance_point_point((proj[0] - point[0], proj[1] - point[1])),
            (line_start[0] + proj[0], line_start[1] + proj[1]),
        )


def intersect_line_line(
    line1_start: Point, line1_end: Point, line2_start: Point, line2_end: Point
) -> Optional[Point]:
    """Find the point where the lines (segments) defined by ``(line1_start,
    line1_end)`` and ``(line2_start, line2_end)`` intersect.  If no
    intersection occurs, ``None`` is returned.

    >>> intersect_line_line((3, 0), (8, 10), (0, 0), (10, 10))
    (6, 6)
    >>> intersect_line_line((0, 0), (10, 10), (3, 0), (8, 10))
    (6, 6)
    >>> intersect_line_line((0, 0), (10, 10), (8, 10), (3, 0))
    (6, 6)
    >>> intersect_line_line((8, 10), (3, 0), (0, 0), (10, 10))
    (6, 6)
    >>> intersect_line_line((0, 0), (0, 10), (3, 0), (8, 10))
    >>> intersect_line_line((0, 0), (0, 10), (3, 0), (3, 10))

    Ticket #168:
    >>> intersect_line_line((478.0, 117.0), (478.0, 166.0), (527.5, 141.5), (336.5, 139.5))
    (478.5, 141.48167539267016)
    >>> intersect_line_line((527.5, 141.5), (336.5, 139.5), (478.0, 117.0), (478.0, 166.0))
    (478.5, 141.48167539267016)

    This is a Python translation of the ``lines_intersect``, C Code from
    Graphics Gems II, Academic Press, Inc. The original routine was written
    by Mukesh Prasad.

    EULA: The Graphics Gems code is copyright-protected. In other words, you
    cannot claim the text of the code as your own and resell it. Using the code
    is permitted in any program, product, or library, non-commercial or
    commercial. Giving credit is not required, though is a nice gesture. The
    code comes as-is, and if there are any flaws or problems with any Gems
    code, nobody involved with Gems - authors, editors, publishers, or
    webmasters - are to be held responsible. Basically, don't be a jerk, and
    remember that anything free comes with no guarantee.
    """
    #
    #   This function computes whether two line segments,
    #   respectively joining the input points (x1,y1) -- (x2,y2)
    #   and the input points (x3,y3) -- (x4,y4) intersect.
    #   If the lines intersect, the output variables x, y are
    #   set to coordinates of the point of intersection.
    #
    #   All values are in integers.  The returned value is rounded
    #   to the nearest integer point.
    #
    #   If non-integral grid points are relevant, the function
    #   can easily be transformed by substituting floating point
    #   calculations instead of integer calculations.
    #
    #   Entry
    #        x1, y1,  x2, y2   Coordinates of endpoints of one segment.
    #        x3, y3,  x4, y4   Coordinates of endpoints of other segment.
    #
    #   Exit
    #        x, y              Coordinates of intersection point.
    #
    #   The value returned by the function is one of:
    #
    #        DONT_INTERSECT    0
    #        DO_INTERSECT      1
    #        COLLINEAR         2
    #
    # Error conditions:
    #
    #     Depending upon the possible ranges, and particularly on 16-bit
    #     computers, care should be taken to protect from overflow.
    #
    #     In the following code, 'long' values have been used for this
    #     purpose, instead of 'int'.
    #

    x1, y1 = line1_start
    x2, y2 = line1_end
    x3, y3 = line2_start
    x4, y4 = line2_end

    # long a1, a2, b1, b2, c1, c2; /* Coefficients of line eqns. */
    # long r1, r2, r3, r4;         /* 'Sign' values */
    # long denom, offset, num;     /* Intermediate values */

    # Compute a1, b1, c1, where line joining points 1 and 2
    # is "a1 x  +  b1 y  +  c1  =  0".

    a1 = y2 - y1
    b1 = x1 - x2
    c1 = x2 * y1 - x1 * y2

    # Compute r3 and r4.

    r3 = a1 * x3 + b1 * y3 + c1
    r4 = a1 * x4 + b1 * y4 + c1

    # Check signs of r3 and r4.  If both point 3 and point 4 lie on
    # same side of line 1, the line segments do not intersect.

    if r3 and r4 and (r3 * r4) >= 0:
        return None  # ( DONT_INTERSECT )

    # Compute a2, b2, c2

    a2 = y4 - y3
    b2 = x3 - x4
    c2 = x4 * y3 - x3 * y4

    # Compute r1 and r2

    r1 = a2 * x1 + b2 * y1 + c2
    r2 = a2 * x2 + b2 * y2 + c2

    # Check signs of r1 and r2.  If both point 1 and point 2 lie
    # on same side of second line segment, the line segments do
    # not intersect.

    if r1 and r2 and (r1 * r2) >= 0:  # SAME_SIGNS( r1, r2 ))
        return None  # ( DONT_INTERSECT )

    # Line segments intersect: compute intersection point.
    # The denom / 2 is to get rounding instead of truncating.  It
    # is added or subtracted to the numerator, depending upon the
    # sign of the numerator.

    denom = a1 * b2 - a2 * b1
    x_num = b1 * c2 - b2 * c1
    y_num = a2 * c1 - a1 * c2
    if not denom:
        return None  # ( COLLINEAR )
    elif isinstance(denom, float):  # denom is float, use normal division
        offset = abs(denom) / 2
        x = ((x_num < 0) and (x_num - offset) or (x_num + offset)) / denom
        y = ((y_num < 0) and (y_num - offset) or (y_num + offset)) / denom
    else:  # denom is int, use integer division
        offset = abs(denom) // 2
        x = ((x_num < 0) and (x_num - offset) or (x_num + offset)) // denom
        y = ((y_num < 0) and (y_num - offset) or (y_num + offset)) // denom
    return x, y


def rectangle_contains(inner: Rect, outer: Rect) -> bool:
    """Returns True if ``inner`` rect is contained in ``outer`` rect."""
    ix, iy, iw, ih = inner
    ox, oy, ow, oh = outer
    return ox <= ix and oy <= iy and ox + ow >= ix + iw and oy + oh >= iy + ih


def rectangle_intersects(recta: Rect, rectb: Rect) -> bool:
    """Return True if ``recta`` and ``rectb`` intersect.

    >>> rectangle_intersects((5,5,20, 20), (10, 10, 1, 1))
    True
    >>> rectangle_intersects((40, 30, 10, 1), (1, 1, 1, 1))
    False
    """
    ax, ay, aw, ah = recta
    bx, by, bw, bh = rectb
    return ax <= bx + bw and ax + aw >= bx and ay <= by + bh and ay + ah >= by


def rectangle_clip(recta: Rect, rectb: Rect) -> Optional[Rect]:
    """Return the clipped rectangle of ``recta`` and ``rectb``. If they do not
    intersect, ``None`` is returned.

    >>> rectangle_clip((0, 0, 20, 20), (10, 10, 20, 20))
    (10, 10, 10, 10)
    """
    ax, ay, aw, ah = recta
    bx, by, bw, bh = rectb
    x = max(ax, bx)
    y = max(ay, by)
    w = min(ax + aw, bx + bw) - x
    h = min(ay + ah, by + bh) - y
    if w < 0 or h < 0:
        return None
    return (x, y, w, h)



from __future__ import annotations

from typing import Collection, List

from gaphas.item import Item
from gaphas.painter.painter import Painter
from gaphas.types import CairoContext


class PainterChain:
    """Chain up a set of painters."""

    def __init__(self) -> None:
        self._painters: List[Painter] = []

    def append(self, painter: Painter) -> PainterChain:
        """Add a painter to the list of painters."""
        self._painters.append(painter)
        return self

    def prepend(self, painter: Painter) -> PainterChain:
        """Add a painter to the beginning of the list of painters."""
        self._painters.insert(0, painter)
        return self

    def paint(self, items: Collection[Item], cairo: CairoContext) -> None:
        """See Painter.paint()."""
        for painter in self._painters:
            painter.paint(items, cairo)



"""A Canvas owns a set of Items and acts as a container for both the items and
a constraint solver.

Connections
===========

Getting Connection Information
==============================
To get connected item to a handle::

    c = canvas.connections.get_connection(handle)
    if c is not None:
        print c.connected
        print c.port
        print c.constraint


To get all connected items (i.e. items on both sides of a line)::

    classes = (i.connected for i in canvas.get_connections(item=line))


To get connecting items (i.e. all lines connected to a class)::

    lines = (c.item for c in canvas.get_connections(connected=item))
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Iterable, Optional

import cairo
from typing_extensions import Protocol

from gaphas import matrix, tree
from gaphas.connections import Connection, Connections
from gaphas.decorators import nonrecursive

if TYPE_CHECKING:
    from gaphas.item import Item
    from gaphas.view.model import View


def instant_cairo_context():
    """A simple Cairo context, not attached to any window."""
    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 0, 0)
    return cairo.Context(surface)


class Canvas:
    """Container class for items."""

    def __init__(self):
        self._tree: tree.Tree[Item] = tree.Tree()
        self._connections = Connections()

        self._registered_views = set()
        self._connections.add_handler(self._on_constraint_solved)

    @property
    def solver(self):
        return self._connections.solver

    @property
    def connections(self) -> Connections:
        return self._connections

    def add(self, item, parent=None, index=None):
        """Add an item to the canvas.

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> len(c._tree.nodes)
        1
        >>> i._canvas is c
        True
        """
        assert item not in self._tree.nodes, f"Adding already added node {item}"

        self._tree.add(item, parent, index)
        self.request_update(item)

    def _remove(self, item):
        """Remove is done in a separate, @observed, method so the undo system
        can restore removed items in the right order."""
        self._tree.remove(item)
        self._connections.disconnect_item(item)
        self._update_views(removed_items=(item,))

    def remove(self, item):
        """Remove item from the canvas.

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> c.remove(i)
        >>> c._tree.nodes
        []
        >>> i._canvas
        """
        for child in reversed(list(self.get_children(item))):
            self.remove(child)
        self._connections.remove_connections_to_item(item)
        self._remove(item)

    def reparent(self, item, parent, index=None):
        """Set new parent for an item."""
        self._tree.move(item, parent, index)

    def get_all_items(self) -> Iterable[Item]:
        """Get a list of all items.

        >>> c = Canvas()
        >>> c.get_all_items()
        []
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> c.get_all_items() # doctest: +ELLIPSIS
        [<gaphas.item.Item ...>]
        """
        return iter(self._tree.nodes)

    def get_root_items(self):
        """Return the root items of the canvas.

        >>> c = Canvas()
        >>> c.get_all_items()
        []
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> ii = item.Item()
        >>> c.add(ii, i)
        >>> c.get_root_items() # doctest: +ELLIPSIS
        [<gaphas.item.Item ...>]
        """
        return self._tree.get_children(None)

    def get_parent(self, item: Item) -> Optional[Item]:
        """See `tree.Tree.get_parent()`.

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> ii = item.Item()
        >>> c.add(ii, i)
        >>> c.get_parent(i)
        >>> c.get_parent(ii) # doctest: +ELLIPSIS
        <gaphas.item.Item ...>
        """
        return self._tree.get_parent(item)

    def get_children(self, item: Optional[Item]) -> Iterable[Item]:
        """See `tree.Tree.get_children()`.

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i = item.Item()
        >>> c.add(i)
        >>> ii = item.Item()
        >>> c.add(ii, i)
        >>> iii = item.Item()
        >>> c.add(iii, ii)
        >>> list(c.get_children(iii))
        []
        >>> list(c.get_children(ii)) # doctest: +ELLIPSIS
        [<gaphas.item.Item ...>]
        >>> list(c.get_children(i)) # doctest: +ELLIPSIS
        [<gaphas.item.Item ...>]
        """
        return self._tree.get_children(item)

    def sort(self, items: Iterable[Item]) -> Iterable[Item]:
        """Sort a list of items in the order in which they are traversed in the
        canvas (Depth first).

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i1 = item.Line()
        >>> c.add(i1)
        >>> i2 = item.Line()
        >>> c.add(i2)
        >>> i3 = item.Line()
        >>> c.add (i3)
        >>> c.update_now((i1, i2, i3)) # ensure items are indexed
        >>> s = c.sort([i2, i3, i1])
        >>> s[0] is i1 and s[1] is i2 and s[2] is i3
        True
        """
        return self._tree.order(items)

    def get_matrix_i2c(self, item: Item) -> matrix.Matrix:
        """Get the Item to Canvas matrix for ``item``.

        item:
            The item who's item-to-canvas transformation matrix should
            be found
        calculate:
            True will allow this function to actually calculate it,
            instead of raising an `AttributeError` when no matrix is
            present yet. Note that out-of-date matrices are not
            recalculated.
        """
        m = item.matrix

        parent = self._tree.get_parent(item)
        if parent is not None:
            m = m.multiply(self.get_matrix_i2c(parent))
        return m

    def request_update(
        self, item: Item, update: bool = True, matrix: bool = True
    ) -> None:
        """Set an update request for the item.

        >>> c = Canvas()
        >>> from gaphas import item
        >>> i = item.Item()
        >>> ii = item.Item()
        >>> c.add(i)
        >>> c.add(ii, i)
        >>> len(c._dirty_items)
        0
        >>> c.update_now((i, ii))
        >>> len(c._dirty_items)
        0
        """
        if update and matrix:
            self._update_views(dirty_items=(item,), dirty_matrix_items=(item,))
        elif update:
            self._update_views(dirty_items=(item,))
        elif matrix:
            self._update_views(dirty_matrix_items=(item,))

    def request_matrix_update(self, item):
        """Schedule only the matrix to be updated."""
        self.request_update(item, update=False, matrix=True)

    @nonrecursive
    def update_now(self, dirty_items, dirty_matrix_items=()):
        """Perform an update of the items that requested an update."""
        try:
            # keep it here, since we need up to date matrices for the solver
            for d in dirty_items:
                d.matrix_i2c.set(*self.get_matrix_i2c(d))
            for d in dirty_matrix_items:
                d.matrix_i2c.set(*self.get_matrix_i2c(d))

            # solve all constraints
            self._connections.solve()

        except Exception as e:
            logging.error("Error while updating canvas", exc_info=e)

    def register_view(self, view: View) -> None:
        """Register a view on this canvas.

        This method is called when setting a canvas on a view and should
        not be called directly from user code.
        """
        self._registered_views.add(view)

    def unregister_view(self, view: View) -> None:
        """Unregister a view on this canvas.

        This method is called when setting a canvas on a view and should
        not be called directly from user code.
        """
        self._registered_views.discard(view)

    def _on_constraint_solved(self, cinfo: Connection) -> None:
        dirty_items = set()
        if cinfo.item:
            dirty_items.add(cinfo.item)
        if cinfo.connected:
            dirty_items.add(cinfo.connected)
        if dirty_items:
            self._update_views(dirty_items)

    def _update_views(self, dirty_items=(), dirty_matrix_items=(), removed_items=()):
        """Send an update notification to all registered views."""
        for v in self._registered_views:
            v.request_update(dirty_items, dirty_matrix_items, removed_items)


class Traversable(Protocol):
    def get_parent(self, item: Item) -> Optional[Item]:
        ...

    def get_children(self, item: Optional[Item]) -> Iterable[Item]:
        ...


def ancestors(canvas: Traversable, item: Item) -> Iterable[Optional[Item]]:
    parent = canvas.get_parent(item)
    while parent:
        yield parent
        parent = canvas.get_parent(parent)


def all_children(canvas: Traversable, item: Optional[Item]) -> Iterable[Item]:
    children = canvas.get_children(item)
    for child in children:
        yield child
        yield from all_children(canvas, child)


# Additional tests in @observed methods
__test__ = {
    "Canvas.add": Canvas.add,
    "Canvas.remove": Canvas.remove,
    "Canvas.request_update": Canvas.request_update,
}



from __future__ import annotations

from typing import TYPE_CHECKING

from gaphas.painter.boundingboxpainter import BoundingBoxPainter
from gaphas.painter.chain import PainterChain
from gaphas.painter.freehand import FreeHandPainter
from gaphas.painter.handlepainter import HandlePainter
from gaphas.painter.itempainter import ItemPainter
from gaphas.painter.painter import Painter

if TYPE_CHECKING:
    from gaphas.view import GtkView


def DefaultPainter(view: GtkView) -> Painter:
    """Default painter, containing item, handle and tool painters."""
    return (
        PainterChain().append(ItemPainter(view.selection)).append(HandlePainter(view))
    )



"""Some Gaphor specific updates to the canvas. This is done by setting the
correct properties on gaphas' modules.

Matrix
------
Small utility class wrapping cairo.Matrix. The `Matrix` class adds
state notification capabilities.
"""

from __future__ import annotations

from typing import Callable, Optional, Set, Tuple

import cairo

MatrixTuple = Tuple[float, float, float, float, float, float]


class Matrix:
    """Matrix wrapper.

    >>> Matrix()
    Matrix(1.0, 0.0, 0.0, 1.0, 0.0, 0.0)
    """

    def __init__(
        self,
        xx: float = 1.0,
        yx: float = 0.0,
        xy: float = 0.0,
        yy: float = 1.0,
        x0: float = 0.0,
        y0: float = 0.0,
        matrix: Optional[cairo.Matrix] = None,
    ) -> None:
        self._matrix = matrix or cairo.Matrix(xx, yx, xy, yy, x0, y0)
        self._handlers: Set[Callable[[Matrix, MatrixTuple], None]] = set()

    def add_handler(
        self,
        handler: Callable[[Matrix, MatrixTuple], None],
    ) -> None:
        self._handlers.add(handler)

    def remove_handler(
        self,
        handler: Callable[[Matrix, MatrixTuple], None],
    ) -> None:
        self._handlers.discard(handler)

    def notify(self, old: MatrixTuple) -> None:
        for handler in self._handlers:
            handler(self, old)

    def invert(self) -> None:
        old: MatrixTuple = self.tuple()
        self._matrix.invert()
        self.notify(old)

    def rotate(self, radians: float) -> None:
        old: MatrixTuple = self.tuple()
        self._matrix.rotate(radians)
        self.notify(old)

    def scale(self, sx: float, sy: float) -> None:
        old = self.tuple()
        self._matrix.scale(sx, sy)
        self.notify(old)

    def translate(self, tx: float, ty: float) -> None:
        old: MatrixTuple = self.tuple()
        self._matrix.translate(tx, ty)
        self.notify(old)

    def set(
        self,
        xx: Optional[float] = None,
        yx: Optional[float] = None,
        xy: Optional[float] = None,
        yy: Optional[float] = None,
        x0: Optional[float] = None,
        y0: Optional[float] = None,
    ) -> None:
        updated = False
        m = self._matrix
        old = self.tuple()
        for name, val in (
            ("xx", xx),
            ("yx", yx),
            ("xy", xy),
            ("yy", yy),
            ("x0", x0),
            ("y0", y0),
        ):
            if val is not None and val != getattr(m, name):
                setattr(m, name, val)
                updated = True
        if updated:
            self.notify(old)

    def multiply(self, m: Matrix) -> Matrix:
        return Matrix(matrix=self._matrix.multiply(m._matrix))

    def transform_distance(self, dx: float, dy: float) -> Tuple[float, float]:
        return self._matrix.transform_distance(dx, dy)  # type: ignore[no-any-return]

    def transform_point(self, x: float, y: float) -> Tuple[float, float]:
        return self._matrix.transform_point(x, y)  # type: ignore[no-any-return]

    def inverse(self) -> Matrix:
        m = Matrix(matrix=cairo.Matrix(*self._matrix))  # type: ignore[misc]
        m.invert()
        return m

    def tuple(self) -> MatrixTuple:
        return tuple(self)  # type: ignore[arg-type, assignment, return-value]

    def to_cairo(self) -> cairo.Matrix:
        return self._matrix

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Matrix):
            return bool(self._matrix == other._matrix)
        else:
            return False

    def __getitem__(self, val: int) -> float:
        return self._matrix[val]  # type: ignore[index,no-any-return]

    def __mul__(self, other: Matrix) -> Matrix:
        return Matrix(matrix=self._matrix * other._matrix)

    def __repr__(self) -> str:
        return f"Matrix{tuple(self._matrix)}"  # type: ignore[arg-type]



from gi.repository import Gtk

from gaphas.view import GtkView


class ZoomData:
    x0: int
    y0: int
    sx: float
    sy: float


def zoom_tool(view: GtkView) -> Gtk.GestureZoom:
    """Create a zoom tool as a Gtk.Gesture.

    Note: we need to keep a reference to this gesture, or else it will be destroyed.
    """
    zoom_data = ZoomData()
    gesture = Gtk.GestureZoom.new(view)
    gesture.set_propagation_phase(Gtk.PropagationPhase.CAPTURE)
    gesture.connect("begin", on_begin, zoom_data)
    gesture.connect("scale-changed", on_scale_changed, zoom_data)
    return gesture


def on_begin(
    gesture: Gtk.GestureZoom,
    sequence: None,
    zoom_data: ZoomData,
) -> None:
    _, zoom_data.x0, zoom_data.y0 = gesture.get_point(sequence)
    view = gesture.get_widget()
    zoom_data.sx = view.matrix[0]
    zoom_data.sy = view.matrix[3]


def on_scale_changed(
    gesture: Gtk.GestureZoom, scale: float, zoom_data: ZoomData
) -> None:
    if zoom_data.sx * scale < 0.2:
        scale = 0.2 / zoom_data.sx
    elif zoom_data.sx * scale > 20.0:
        scale = 20.0 / zoom_data.sx

    view = gesture.get_widget()
    m = view.matrix
    sx = m[0]
    sy = m[3]
    ox = (m[4] - zoom_data.x0) / sx
    oy = (m[5] - zoom_data.y0) / sy
    dsx = zoom_data.sx * scale / sx
    dsy = zoom_data.sy * scale / sy
    m.translate(-ox, -oy)
    m.scale(dsx, dsy)
    m.translate(+ox, +oy)
    view.request_update((), view.model.get_all_items())



"""Basic items."""
from __future__ import annotations

from dataclasses import dataclass
from math import atan2
from typing import TYPE_CHECKING, Iterable, List, Sequence

from typing_extensions import Protocol, runtime_checkable

from gaphas.connector import Handle, LinePort, Port
from gaphas.constraint import Constraint, EqualsConstraint, constraint
from gaphas.geometry import distance_line_point, distance_rectangle_point
from gaphas.matrix import Matrix
from gaphas.solver import REQUIRED, VERY_STRONG, variable
from gaphas.types import CairoContext

if TYPE_CHECKING:
    from gaphas.connections import Connections


@dataclass(frozen=True)
class DrawContext:
    cairo: CairoContext
    selected: bool
    focused: bool
    hovered: bool


@runtime_checkable
class Item(Protocol):
    """This protocol should be implemented by model items.

    All items that are rendered on a view.
    """

    @property
    def matrix(self) -> Matrix:
        """The "local", item-to-parent matrix."""

    @property
    def matrix_i2c(self) -> Matrix:
        """Matrix from item to toplevel."""

    def handles(self) -> Sequence[Handle]:
        """Return a list of handles owned by the item."""

    def ports(self) -> Sequence[Port]:
        """Return list of ports owned by the item."""

    def point(self, x: float, y: float) -> float:
        """Get the distance from a point (``x``, ``y``) to the item.

        ``x`` and ``y`` are in item coordinates.

        A distance of 0 means the point is on the item.
        """

    def draw(self, context: DrawContext) -> None:
        """Render the item to a canvas view. Context contains the following
        attributes:

        * `cairo`: the CairoContext use this one to draw
        * `selected`, `focused`, `hovered`: view state of items
          (True/False)
        """


def matrix_i2i(from_item: Item, to_item: Item) -> Matrix:
    i2c = from_item.matrix_i2c
    c2i = to_item.matrix_i2c.inverse()
    return i2c.multiply(c2i)


class Matrices:
    def __init__(self, **kwargs: object) -> None:
        super().__init__(**kwargs)  # type: ignore[call-arg]
        self._matrix = Matrix()
        self._matrix_i2c = Matrix()

    @property
    def matrix(self) -> Matrix:
        return self._matrix

    @property
    def matrix_i2c(self) -> Matrix:
        return self._matrix_i2c


[NW, NE, SE, SW] = list(range(4))


class Element(Matrices):
    """An Element has 4 handles (for a start)::

    NW +---+ NE    |   | SW +---+ SE
    """

    min_width = variable(strength=REQUIRED, varname="_min_width")
    min_height = variable(strength=REQUIRED, varname="_min_height")

    def __init__(
        self,
        connections: Connections,
        width: float = 10,
        height: float = 10,
        **kwargs: object
    ) -> None:
        super().__init__(**kwargs)
        self._handles = [h(strength=VERY_STRONG) for h in [Handle] * 4]

        handles = self._handles
        h_nw = handles[NW]
        h_ne = handles[NE]
        h_sw = handles[SW]
        h_se = handles[SE]

        # edge of element define default element ports
        self._ports = [
            LinePort(h_nw.pos, h_ne.pos),
            LinePort(h_ne.pos, h_se.pos),
            LinePort(h_se.pos, h_sw.pos),
            LinePort(h_sw.pos, h_nw.pos),
        ]

        # initialize min_x variables
        self.min_width, self.min_height = 10, 10

        add = connections.add_constraint
        add(self, constraint(horizontal=(h_nw.pos, h_ne.pos)))
        add(self, constraint(horizontal=(h_sw.pos, h_se.pos)))
        add(self, constraint(vertical=(h_nw.pos, h_sw.pos)))
        add(self, constraint(vertical=(h_ne.pos, h_se.pos)))

        # create minimal size constraints
        add(self, constraint(left_of=(h_nw.pos, h_se.pos), delta=self.min_width))
        add(self, constraint(above=(h_nw.pos, h_se.pos), delta=self.min_height))

        self.width = width
        self.height = height

        # Trigger solver to honour width/height by SE handle pos
        self._handles[SE].pos.x.dirty()
        self._handles[SE].pos.y.dirty()

    @property
    def width(self) -> float:
        """Width of the box, calculated as the distance from the left and right
        handle."""
        h = self._handles
        return float(h[SE].pos.x) - float(h[NW].pos.x)

    @width.setter
    def width(self, width: float) -> None:
        """
        >>> b=Element()
        >>> b.width = 20
        >>> b.width
        20.0
        >>> b._handles[NW].pos.x
        Variable(0, 40)
        >>> b._handles[SE].pos.x
        Variable(20, 40)
        """
        h = self._handles
        h[SE].pos.x = h[NE].pos.x = h[NW].pos.x + width

    @property
    def height(self) -> float:
        """Height."""
        h = self._handles
        return float(h[SE].pos.y) - float(h[NW].pos.y)

    @height.setter
    def height(self, height: float) -> None:
        """
        >>> b=Element()
        >>> b.height = 20
        >>> b.height
        20.0
        >>> b.height = 2
        >>> b.height
        2.0
        >>> b._handles[NW].pos.y
        Variable(0, 40)
        >>> b._handles[SE].pos.y
        Variable(2, 40)
        """
        h = self._handles
        h[SE].pos.y = h[SW].pos.y = h[NW].pos.y + height

    def handles(self) -> Sequence[Handle]:
        """Return a list of handles owned by the item."""
        return self._handles

    def ports(self) -> Sequence[Port]:
        """Return list of ports."""
        return self._ports

    def point(self, x: float, y: float) -> float:
        """Distance from the point (x, y) to the item.

        >>> e = Element()
        >>> e.point(20, 10)
        10.0
        """
        h = self._handles
        x0, y0 = h[NW].pos
        x1, y1 = h[SE].pos
        return distance_rectangle_point((x0, y0, x1 - x0, y1 - y0), (x, y))

    def draw(self, context: DrawContext) -> None:
        pass


def create_orthogonal_constraints(
    handles: Sequence[Handle], horizontal: bool
) -> Iterable[Constraint]:
    rest = 1 if horizontal else 0
    for pos, (h0, h1) in enumerate(zip(handles, handles[1:])):
        p0 = h0.pos
        p1 = h1.pos
        if pos % 2 == rest:
            yield EqualsConstraint(a=p0.x, b=p1.x)
        else:
            yield EqualsConstraint(a=p0.y, b=p1.y)


class Line(Matrices):
    """A Line item.

    Properties:
     - fuzziness (0.0..n): an extra margin that should be taken into
         account when calculating the distance from the line (using
         point()).
     - orthogonal (bool): whether or not the line should be
         orthogonal (only straight angles)
     - horizontal: first line segment is horizontal
     - line_width: width of the line to be drawn

    This line also supports arrow heads on both the begin and end of
    the line. These are drawn with the methods draw_head(context) and
    draw_tail(context). The coordinate system is altered so the
    methods do not have to know about the angle of the line segment
    (e.g. drawing a line from (10, 10) via (0, 0) to (10, -10) will
    draw an arrow point).
    """

    def __init__(self, connections: Connections, **kwargs: object) -> None:
        super().__init__(**kwargs)
        self._connections = connections
        self._handles = [Handle(connectable=True), Handle((10, 10), connectable=True)]
        self._ports: List[Port] = []
        self._update_ports()

        self._line_width = 2.0
        self._fuzziness = 0.0
        self._orthogonal_constraints: List[Constraint] = []
        self._horizontal = False

    @property
    def head(self) -> Handle:
        return self._handles[0]

    @property
    def tail(self) -> Handle:
        return self._handles[-1]

    def _set_line_width(self, line_width: float) -> None:
        self._line_width = line_width

    line_width = property(lambda s: s._line_width, _set_line_width)

    def _set_fuzziness(self, fuzziness: float) -> None:
        self._fuzziness = fuzziness

    fuzziness = property(lambda s: s._fuzziness, _set_fuzziness)

    def update_orthogonal_constraints(self, orthogonal: bool) -> None:
        """Update the constraints required to maintain the orthogonal line.

        The actual constraints attribute (``_orthogonal_constraints``)
        is observed, so the undo system will update the contents
        properly
        """
        for c in self._orthogonal_constraints:
            self._connections.remove_constraint(self, c)
        del self._orthogonal_constraints[:]

        if not orthogonal:
            return

        add = self._connections.add_constraint
        cons = [
            add(self, c)
            for c in create_orthogonal_constraints(self._handles, self._horizontal)
        ]
        self._set_orthogonal_constraints(cons)

    def _set_orthogonal_constraints(
        self, orthogonal_constraints: List[Constraint]
    ) -> None:
        """Setter for the constraints maintained.

        Required for the undo system.
        """
        self._orthogonal_constraints = orthogonal_constraints

    def _set_orthogonal(self, orthogonal: bool) -> None:
        """
        >>> a = Line()
        >>> a.orthogonal
        False
        """
        if orthogonal and len(self.handles()) < 3:
            raise ValueError("Can't set orthogonal line with less than 3 handles")
        self.update_orthogonal_constraints(orthogonal)

    orthogonal = property(lambda s: bool(s._orthogonal_constraints), _set_orthogonal)

    def _inner_set_horizontal(self, horizontal: bool) -> None:
        self._horizontal = horizontal

    def _set_horizontal(self, horizontal: bool) -> None:
        """
        >>> line = Line()
        >>> line.horizontal
        False
        >>> line.horizontal = False
        >>> line.horizontal
        False
        """
        self._inner_set_horizontal(horizontal)
        self.update_orthogonal_constraints(self.orthogonal)

    horizontal = property(lambda s: s._horizontal, _set_horizontal)

    def insert_handle(self, index: int, handle: Handle) -> None:
        self._handles.insert(index, handle)

    def remove_handle(self, handle: Handle) -> None:
        self._handles.remove(handle)

    def insert_port(self, index: int, port: Port) -> None:
        self._ports.insert(index, port)

    def remove_port(self, port: Port) -> None:
        self._ports.remove(port)

    def _update_ports(self) -> None:
        """Update line ports.

        This destroys all previously created ports and should only be
        used when initializing the line.
        """
        assert len(self._handles) >= 2, "Not enough segments"
        self._ports = []
        handles = self._handles
        for h1, h2 in zip(handles[:-1], handles[1:]):
            self._ports.append(LinePort(h1.pos, h2.pos))

    def opposite(self, handle: Handle) -> Handle:
        """Given the handle of one end of the line, return the other end."""
        handles = self._handles
        if handle is handles[0]:
            return handles[-1]
        elif handle is handles[-1]:
            return handles[0]
        else:
            raise KeyError("Handle is not an end handle")

    def handles(self) -> Sequence[Handle]:
        """Return a list of handles owned by the item."""
        return self._handles

    def ports(self) -> Sequence[Port]:
        """Return list of ports."""
        return self._ports

    def point(self, x: float, y: float) -> float:
        """
        >>> a = Line()
        >>> a.handles()[1].pos = 25, 5
        >>> a._handles.append(a._create_handle((30, 30)))
        >>> a.point(-1, 0)
        1.0
        >>> f"{a.point(5, 4):.3f}"
        '2.942'
        >>> f"{a.point(29, 29):.3f}"
        '0.784'
        """
        hpos = [h.pos for h in self._handles]

        distance, _point = min(
            map(distance_line_point, hpos[:-1], hpos[1:], [(x, y)] * (len(hpos) - 1))  # type: ignore[arg-type]
        )
        fuzziness: float = self.fuzziness
        return max(0.0, distance - fuzziness)

    def draw_head(self, context: DrawContext) -> None:
        """Default head drawer: move cursor to the first handle."""
        context.cairo.move_to(0, 0)

    def draw_tail(self, context: DrawContext) -> None:
        """Default tail drawer: draw line to the last handle."""
        context.cairo.line_to(0, 0)

    def draw(self, context: DrawContext) -> None:
        """Draw the line itself.

        See Item.draw(context).
        """

        def draw_line_end(pos, angle, draw):  # type: ignore[no-untyped-def]
            cr = context.cairo
            cr.save()
            try:
                cr.translate(*pos)
                cr.rotate(angle)
                draw(context)
            finally:
                cr.restore()

        cr = context.cairo
        cr.set_line_width(self.line_width)

        h0, h1 = self._handles[:2]
        p0, p1 = h0.pos, h1.pos
        head_angle = atan2(p1.y - p0.y, p1.x - p0.x)  # type: ignore[assignment]
        draw_line_end(self._handles[0].pos, head_angle, self.draw_head)

        for h in self._handles[1:-1]:
            cr.line_to(*h.pos)

        h1, h0 = self._handles[-2:]
        p1, p0 = h1.pos, h0.pos
        tail_angle = atan2(p1.y - p0.y, p1.x - p0.x)  # type: ignore[assignment]
        draw_line_end(self._handles[-1].pos, tail_angle, self.draw_tail)
        cr.stroke()


__test__ = {"Line._set_orthogonal": Line._set_orthogonal}



from gi.repository import Gtk


def view_focus_tool(view):
    """This little tool ensures the view grabs focus when a mouse press or
    touch event happens."""
    gesture = Gtk.GestureSingle(widget=view)
    gesture.connect("begin", on_begin)
    return gesture


def on_begin(gesture, sequence):
    view = gesture.get_widget()
    if not view.is_focus():
        view.grab_focus()



from typing import Collection

from gi.repository import Gtk

from gaphas.item import Item
from gaphas.types import CairoContext


class RubberbandState:
    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        self.x0 = self.y0 = self.x1 = self.y1 = 0


class RubberbandPainter:
    """The rubberband painter should be used in conjunction with the rubberband
    tool.

    ``RubberbandState`` should be shared between the two.
    """

    def __init__(self, rubberband_state: RubberbandState) -> None:
        self.rubberband_state = rubberband_state

    def paint(self, items: Collection[Item], cairo: CairoContext) -> None:
        data = self.rubberband_state
        x0, y0, x1, y1 = data.x0, data.y0, data.x1, data.y1
        if x0 != x1 or y0 != y1:
            cairo.rectangle(min(x0, x1), min(y0, y1), abs(x1 - x0), abs(y1 - y0))
            cairo.set_source_rgba(0.9, 0.9, 0.9, 0.3)
            cairo.fill_preserve()
            cairo.set_line_width(2.0)
            cairo.set_dash((7.0, 5.0), 0)
            cairo.set_source_rgba(0.5, 0.5, 0.7, 0.7)
            cairo.stroke()


def rubberband_tool(view, rubberband_state):
    """Rubberband selection tool.

    Should be used in conjunction with ``RubberbandPainter``.
    """
    gesture = Gtk.GestureDrag.new(view)
    gesture.connect("drag-begin", on_drag_begin, rubberband_state)
    gesture.connect("drag-update", on_drag_update, rubberband_state)
    gesture.connect("drag-end", on_drag_end, rubberband_state)
    return gesture


def on_drag_begin(gesture, start_x, start_y, rubberband_state):
    if gesture.set_state(Gtk.EventSequenceState.CLAIMED):
        rubberband_state.x0 = rubberband_state.x1 = start_x
        rubberband_state.y0 = rubberband_state.y1 = start_y


def on_drag_update(gesture, offset_x, offset_y, rubberband_state):
    rubberband_state.x1 = rubberband_state.x0 + offset_x
    rubberband_state.y1 = rubberband_state.y0 + offset_y
    view = gesture.get_widget()
    view.queue_redraw()


def on_drag_end(gesture, offset_x, offset_y, rubberband_state):
    view = gesture.get_widget()
    x0 = rubberband_state.x0
    y0 = rubberband_state.y0
    x1 = x0 + offset_x
    y1 = y0 + offset_y
    items = view.get_items_in_rectangle(
        (min(x0, x1), min(y0, y1), abs(x1 - x0), abs(y1 - y0)), contain=True
    )
    view.selection.select_items(*items)
    rubberband_state.reset()
    view.queue_redraw()



from typing import List, Optional, Sequence, SupportsFloat, Tuple, TypeVar

import cairo
from typing_extensions import Protocol

# A primitive position, tuple ``(x, y)``
# Pos = Tuple[Union[float, SupportsFloat], Union[float, SupportsFloat]]
Pos = Tuple[float, float]
SupportsFloatPos = Tuple[SupportsFloat, SupportsFloat]

GetT = TypeVar("GetT", covariant=True)
SetT = TypeVar("SetT", contravariant=True)


class TypedProperty(Protocol[GetT, SetT]):
    def __get__(self, obj: object, type: Optional[type] = ...) -> GetT:
        ...

    def __set__(self, obj: object, value: SetT) -> None:
        ...

    def __delete__(self, obj: object) -> None:
        ...


class CairoContext(Protocol):
    def append_path(self, path: cairo.Path) -> None:
        ...

    def arc(
        self, xc: float, yc: float, radius: float, angle1: float, angle2: float
    ) -> None:
        ...

    def arc_negative(
        self, xc: float, yc: float, radius: float, angle1: float, angle2: float
    ) -> None:
        ...

    def clip(self) -> None:
        ...

    def clip_extents(self) -> Tuple[float, float, float, float]:
        ...

    def clip_preserve(self) -> None:
        ...

    def close_path(self) -> None:
        ...

    def copy_clip_rectangle_list(self) -> List[cairo.Rectangle]:
        ...

    def copy_path(self) -> cairo.Path:
        ...

    def copy_path_flat(self) -> cairo.Path:
        ...

    def curve_to(
        self, x1: float, y1: float, x2: float, y2: float, x3: float, y3: float
    ) -> None:
        ...

    def device_to_user(self, x: float, y: float) -> Tuple[float, float]:
        ...

    def device_to_user_distance(self, dx: float, dy: float) -> Tuple[float, float]:
        ...

    def fill(self) -> None:
        ...

    def fill_extents(self) -> Tuple[float, float, float, float]:
        ...

    def fill_preserve(self) -> None:
        ...

    def font_extents(self) -> Tuple[float, float, float, float, float]:
        ...

    def get_antialias(self) -> cairo.Antialias:
        ...

    def get_current_point(self) -> Tuple[float, float]:
        ...

    def get_dash(self) -> Tuple[List[float], float]:
        ...

    def get_dash_count(self) -> int:
        ...

    def get_fill_rule(self) -> cairo.FillRule:
        ...

    def get_group_target(self) -> cairo.Surface:
        ...

    def get_line_cap(self) -> cairo.LineCap:
        ...

    def get_line_join(self) -> cairo.LineJoin:
        ...

    def get_line_width(self) -> float:
        ...

    def get_matrix(self) -> cairo.Matrix:
        ...

    def get_miter_limit(self) -> float:
        ...

    def get_operator(self) -> cairo.Operator:
        ...

    def get_source(self) -> cairo.Pattern:
        ...

    def has_current_point(self) -> bool:
        ...

    def identity_matrix(self) -> None:
        ...

    def in_clip(self, x: float, y: float) -> bool:
        ...

    def in_fill(self, x: float, y: float) -> bool:
        ...

    def in_stroke(self, x: float, y: float) -> bool:
        ...

    def line_to(self, x: float, y: float) -> None:
        ...

    def mask(self, pattern: cairo.Pattern) -> None:
        ...

    def mask_surface(
        self, surface: cairo.Surface, x: float = 0.0, y: float = 0.0
    ) -> None:
        ...

    def move_to(self, x: float, y: float) -> None:
        ...

    def new_path(self) -> None:
        ...

    def new_sub_path(self) -> None:
        ...

    def paint(self) -> None:
        ...

    def paint_with_alpha(self, alpha: float) -> None:
        ...

    def path_extents(self) -> Tuple[float, float, float, float]:
        ...

    def pop_group(self) -> cairo.SurfacePattern:
        ...

    def pop_group_to_source(self) -> None:
        ...

    def push_group(self) -> None:
        ...

    def push_group_with_content(self, content: cairo.Content) -> None:
        ...

    def rectangle(self, x: float, y: float, width: float, height: float) -> None:
        ...

    def rel_curve_to(
        self, dx1: float, dy1: float, dx2: float, dy2: float, dx3: float, dy4: float
    ) -> None:
        ...

    def rel_line_to(self, dx: float, dy: float) -> None:
        ...

    def rel_move_to(self, dx: float, dy: float) -> None:
        ...

    def reset_clip(self) -> None:
        ...

    def restore(self) -> None:
        ...

    def rotate(self, angle: float) -> None:
        ...

    def save(self) -> None:
        ...

    def scale(self, sx: float, sy: float) -> None:
        ...

    def set_antialias(self, antialias: cairo.Antialias) -> None:
        ...

    def set_dash(self, dashes: Sequence[float], offset: float = 0) -> None:
        ...

    def set_fill_rule(self, fill_rule: cairo.FillRule) -> None:
        ...

    def set_font_matrix(self, matrix: cairo.Matrix) -> None:
        ...

    def set_line_cap(self, line_cap: cairo.LineCap) -> None:
        ...

    def set_line_join(self, line_join: cairo.LineJoin) -> None:
        ...

    def set_line_width(self, width: float) -> None:
        ...

    def set_matrix(self, matrix: cairo.Matrix) -> None:
        ...

    def set_miter_limit(self, limit: float) -> None:
        ...

    def set_operator(self, op: cairo.Operator) -> None:
        ...

    def set_scaled_font(self, scaled_font: cairo.ScaledFont) -> None:
        ...

    def set_source(self, source: cairo.Pattern) -> None:
        ...

    def set_source_rgb(self, red: float, green: float, blue: float) -> None:
        ...

    def set_source_rgba(
        self, red: float, green: float, blue: float, alpha: float = 1.0
    ) -> None:
        ...

    def set_tolerance(self, tolerance: float) -> None:
        ...

    def stroke(self) -> None:
        ...

    def stroke_extents(self) -> Tuple[float, float, float, float]:
        ...

    def stroke_preserve(self) -> None:
        ...

    def transform(self, matrix: cairo.Matrix) -> None:
        ...

    def translate(self, tx: float, ty: float) -> None:
        ...

    def user_to_device(self, x: float, y: float) -> Tuple[float, float]:
        ...

    def user_to_device_distance(self, dx: float, dy: float) -> Tuple[float, float]:
        ...



from typing import Optional

from gi.repository import Gtk

from gaphas.item import Item
from gaphas.tool.itemtool import handle_at_point, item_at_point
from gaphas.types import Pos
from gaphas.view import GtkView


def hover_tool(view: GtkView) -> Gtk.EventController:
    """Highlight the currenly hovered item."""
    ctrl = Gtk.EventControllerMotion.new(view)
    ctrl.connect("motion", on_motion)
    return ctrl


def on_motion(ctrl, x, y):
    view = ctrl.get_widget()
    view.selection.hovered_item = find_item_at_point(view, (x, y))


def find_item_at_point(view: GtkView, pos: Pos) -> Optional[Item]:
    item, handle = handle_at_point(view, pos)
    return item or item_at_point(view, pos)



"""Allow for easily adding segments to lines."""
from functools import singledispatch
from typing import Optional

from cairo import ANTIALIAS_NONE
from gi.repository import Gtk

from gaphas.aspect.handlemove import HandleMove, ItemHandleMove
from gaphas.connector import Handle, LinePort
from gaphas.geometry import distance_line_point, distance_point_point_fast
from gaphas.item import Line, matrix_i2i
from gaphas.solver import WEAK
from gaphas.tool.itemtool import MoveType, item_at_point
from gaphas.view import Selection
from gaphas.view.model import Model


@singledispatch
class Segment:
    def __init__(self, item, model):
        raise TypeError

    def split_segment(self, segment, count=2):
        ...

    def split(self, pos):
        ...

    def merge_segment(self, segment, count=2):
        ...


@Segment.register(Line)  # type: ignore
class LineSegment:
    def __init__(self, item: Line, model: Model):
        self.item = item
        self.model = model

    def split(self, pos):
        item = self.item
        handles = item.handles()
        x, y = item.matrix_i2c.inverse().transform_point(*pos)
        for h1, h2 in zip(handles, handles[1:]):
            xp = (h1.pos.x + h2.pos.x) / 2
            yp = (h1.pos.y + h2.pos.y) / 2
            if distance_point_point_fast((x, y), (xp, yp)) <= 4:
                segment = handles.index(h1)
                handles, ports = self.split_segment(segment)
                return handles and handles[0]

    def split_segment(self, segment, count=2):
        """Split one item segment into ``count`` equal pieces.

        def split_segment(self, segment, count=2):
        Two lists are returned

          - list of created handles
          - list of created ports

          :Parameters:
           segment
              Segment number to split (starting from zero).
           count
              Amount of new segments to be created (minimum 2).
        """
        item = self.item
        if segment < 0 or segment >= len(item.ports()):
            raise ValueError("Incorrect segment")
        if count < 2:
            raise ValueError("Incorrect count of segments")

        def do_split(segment, count):
            handles = item.handles()
            p0 = handles[segment].pos
            p1 = handles[segment + 1].pos
            dx, dy = p1.x - p0.x, p1.y - p0.y
            new_h = Handle((p0.x + dx / count, p0.y + dy / count), strength=WEAK)
            item.insert_handle(segment + 1, new_h)

            p0 = LinePort(p0, new_h.pos)
            p1 = LinePort(new_h.pos, p1)
            item.remove_port(item.ports()[segment])
            item.insert_port(segment, p0)
            item.insert_port(segment + 1, p1)

            if count > 2:
                do_split(segment + 1, count - 1)

        do_split(segment, count)

        # force orthogonal constraints to be recreated
        item.update_orthogonal_constraints(item.orthogonal)

        self.recreate_constraints()

        self.model.request_update(item)
        handles = item.handles()[segment + 1 : segment + count]
        ports = item.ports()[segment : segment + count - 1]
        return handles, ports

    def merge_segment(self, segment, count=2):
        """Merge two (or more) item segments.

        Tuple of two lists is returned, list of deleted handles and
        list of deleted ports.

        :Parameters:
         segment
            Segment number to start merging from (starting from zero).
         count
            Amount of segments to be merged (minimum 2).
        """
        item = self.item
        if len(item.ports()) < 2:
            raise ValueError("Cannot merge item with one segment")
        if segment < 0 or segment >= len(item.ports()):
            raise ValueError("Incorrect segment")
        if count < 2 or segment + count > len(item.ports()):
            raise ValueError("Incorrect count of segments")

        # remove handle and ports which share position with handle
        deleted_handles = item.handles()[segment + 1 : segment + count]
        deleted_ports = item.ports()[segment : segment + count]
        for h in deleted_handles:
            item.remove_handle(h)
        for p in deleted_ports:
            item.remove_port(p)

        # create new port, which replaces old ports destroyed due to
        # deleted handle
        p1 = item.handles()[segment].pos
        p2 = item.handles()[segment + 1].pos
        port = LinePort(p1, p2)
        item.insert_port(segment, port)

        # force orthogonal constraints to be recreated
        item.update_orthogonal_constraints(item.orthogonal)

        self.recreate_constraints()
        self.model.request_update(item)

        return deleted_handles, deleted_ports

    def recreate_constraints(self):
        """Create connection constraints between connecting lines and an item.

        :Parameters:
         connected
            Connected item.
        """
        connected = self.item
        model = self.model

        def find_port(line, handle, item):
            """Glue to the closest item on the canvas.

            If the item can connect, it returns a port.
            """
            pos = matrix_i2i(line, item).transform_point(*handle.pos)
            port = None
            max_dist = 10e6
            for p in item.ports():
                pg, d = p.glue(pos)
                if d >= max_dist:
                    continue
                port = p
                max_dist = d

            return port

        for cinfo in list(model.connections.get_connections(connected=connected)):
            item, handle = cinfo.item, cinfo.handle
            port = find_port(item, handle, connected)

            constraint = port.constraint(item, handle, connected)

            model.connections.reconnect_item(item, handle, port, constraint=constraint)


class SegmentState:
    moving: Optional[MoveType]

    def __init__(self):
        self.reset()

    def reset(self):
        self.moving = None


def segment_tool(view):
    gesture = Gtk.GestureDrag.new(view)
    segment_state = SegmentState()
    gesture.connect("drag-begin", on_drag_begin, segment_state)
    gesture.connect("drag-update", on_drag_update, segment_state)
    gesture.connect("drag-end", on_drag_end, segment_state)
    return gesture


def on_drag_begin(gesture, start_x, start_y, segment_state):
    view = gesture.get_widget()
    pos = (start_x, start_y)
    item = item_at_point(view, pos)
    handle = item and maybe_split_segment(view, item, pos)
    if handle:
        segment_state.moving = HandleMove(item, handle, view)
        gesture.set_state(Gtk.EventSequenceState.CLAIMED)
    else:
        gesture.set_state(Gtk.EventSequenceState.DENIED)


def on_drag_update(gesture, offset_x, offset_y, segment_state):
    _, x, y = gesture.get_start_point()
    segment_state.moving.move((x + offset_x, y + offset_y))


def on_drag_end(gesture, offset_x, offset_y, segment_state):
    if not segment_state.moving:
        return

    _, x, y = gesture.get_start_point()
    segment_state.moving.stop_move((x + offset_x, y + offset_y))
    segment_state.reset()


@HandleMove.register(Line)
class LineHandleMove(ItemHandleMove):
    def stop_move(self, pos):
        super().stop_move(pos)
        maybe_merge_segments(self.view, self.item, self.handle)


def maybe_split_segment(view, item, pos):
    item = view.selection.hovered_item
    handle = None
    if item is view.selection.focused_item:
        try:
            segment = Segment(item, view.model)
        except TypeError:
            pass
        else:
            cpos = view.matrix.inverse().transform_point(*pos)
            handle = segment.split(cpos)
    return handle


def maybe_merge_segments(view, item, handle):
    handles = item.handles()

    # don't merge using first or last handle
    if handles[0] is handle or handles[-1] is handle:
        return True

    handle_index = handles.index(handle)
    segment = handle_index - 1

    # cannot merge starting from last segment
    if segment == len(item.ports()) - 1:
        segment = -1
    assert segment >= 0 and segment < len(item.ports()) - 1

    before = handles[handle_index - 1]
    after = handles[handle_index + 1]
    d, p = distance_line_point(before.pos, after.pos, handle.pos)

    if d < 2:
        Segment(item, view.model).merge_segment(segment)

    if handle:
        view.model.request_update(item)


class LineSegmentPainter:
    """This painter draws pseudo-handles on gaphas.item.Line objects. Each line
    can be split by dragging those points, which will result in a new handle.

    ConnectHandleTool take care of performing the user interaction
    required for this feature.
    """

    def __init__(self, selection: Selection):
        self.selection = selection

    def paint(self, _items, cairo):
        selection = self.selection
        item = selection.hovered_item
        if isinstance(item, Line) and item is selection.focused_item:
            h = item.handles()
            for h1, h2 in zip(h[:-1], h[1:]):
                p1, p2 = h1.pos, h2.pos
                cx = (p1.x + p2.x) / 2
                cy = (p1.y + p2.y) / 2
                vx, vy = cairo.user_to_device(*item.matrix_i2c.transform_point(cx, cy))
                cairo.save()
                cairo.set_antialias(ANTIALIAS_NONE)
                cairo.identity_matrix()
                cairo.translate(vx, vy)
                cairo.rectangle(-3, -3, 6, 6)
                cairo.set_source_rgba(0, 0.5, 0, 0.4)
                cairo.fill_preserve()
                cairo.set_source_rgba(0.25, 0.25, 0.25, 0.6)
                cairo.set_line_width(1)
                cairo.stroke()
                cairo.restore()



"""Helper functions and classes for Cairo (drawing engine used by the
canvas)."""
from math import pi

import cairo


def text_extents(cr, text, font=None, multiline=False, padding=1):
    """Simple way to determine the size of a piece of text."""
    if not text:
        return 0, 0
    if font:
        cr.save()
        text_set_font(cr, font)

    if multiline:
        width, height = 0, 0
        for line in text.split("\n"):
            x_bear, y_bear, w, h, x_adv, y_adv = cr.text_extents(line)
            width = max(width, w)
            height += h + padding
    else:
        x_bear, y_bear, width, height, x_adv, y_adv = cr.text_extents(text)
        # width, height = width + x_bearing, height + y_bearing

    if font:
        cr.restore()
    return width, height


def text_center(cr, x, y, text):
    text_align(cr, x, y, text, align_x=0, align_y=0)


def text_align(cr, x, y, text, align_x=0, align_y=0, padding_x=0, padding_y=0):
    """Draw text relative to (x, y).

    x, y - coordinates
    text - text to print (utf8)
    align_x - -1 (top), 0 (middle), 1 (bottom)
    align_y - -1 (left), 0 (center), 1 (right)
    padding_x - padding (extra offset), always > 0
    padding_y - padding (extra offset), always > 0
    """
    if not text:
        return

    x_bear, y_bear, w, h, x_adv, y_adv = cr.text_extents(text)
    if align_x == 0:
        x = 0.5 - (w / 2 + x_bear) + x
    elif align_x < 0:
        x = -(w + x_bear) + x - padding_x
    else:
        x = x + padding_x
    if align_y == 0:
        y = 0.5 - (h / 2 + y_bear) + y
    elif align_y < 0:
        y = -(h + y_bear) + y - padding_y
    else:
        y = -y_bear + y + padding_y
    cr.move_to(x, y)
    cr.show_text(text)


def text_multiline(cr, x, y, text, padding=1):
    """Draw a string of text with embedded newlines.

    cr - cairo context
    x - leftmost x
    y - topmost y
    text - text to draw
    padding - additional padding between lines.
    """
    if not text:
        return
    # cr.move_to(x, y)
    for line in text.split("\n"):
        x_bear, y_bear, w, h, x_adv, y_adv = cr.text_extents(text)
        y += h
        cr.move_to(x, y)
        cr.show_text(line)


def text_underline(cr, x, y, text, offset=1.5):
    """Draw text with underline."""
    x_bear, y_bear, w, h, x_adv, y_adv = cr.text_extents(text)
    cr.move_to(x, y - y_bear)
    cr.show_text(text)
    cr.move_to(x, y - y_bear + offset)
    cr.set_line_width(1.0)
    cr.rel_line_to(x_adv, 0)
    cr.stroke()


def text_set_font(cr, font):
    """Set the font from a string.

    E.g. 'sans 10' or 'sans italic bold 12' only restriction is that the
    font name should be the first option and the font size as last
    argument
    """
    font = font.split()
    cr.select_font_face(
        font[0],
        "italic" in font and cairo.FONT_SLANT_ITALIC or cairo.FONT_SLANT_NORMAL,
        "bold" in font and cairo.FONT_WEIGHT_BOLD or cairo.FONT_WEIGHT_NORMAL,
    )
    cr.set_font_size(float(font[-1]))


def path_ellipse(cr, x, y, width, height, angle=0):
    """Draw an ellipse.

    x      - center x
    y      - center y
    width  - width of ellipse  (in x direction when angle=0)
    height - height of ellipse (in y direction when angle=0)
    angle  - angle in radians to rotate, clockwise
    """
    cr.save()
    cr.translate(x, y)
    cr.rotate(angle)
    cr.scale(width / 2.0, height / 2.0)
    cr.move_to(1.0, 0.0)
    cr.arc(0.0, 0.0, 1.0, 0.0, 2.0 * pi)
    cr.restore()



from __future__ import annotations

import functools
import operator
from typing import Collection, Dict, Optional

from gaphas.geometry import Rectangle
from gaphas.item import Item
from gaphas.painter.painter import ItemPainterType
from gaphas.types import CairoContext


class CairoBoundingBoxContext:
    """Delegate all calls to the wrapped CairoBoundingBoxContext, intercept
    ``stroke()``, ``fill()`` and a few others so the bounding box of the item
    involved can be calculated."""

    def __init__(self, cr):
        self.cr = cr
        self._bounds: Optional[Rectangle] = None  # a Rectangle object

    def __getattr__(self, key):
        return getattr(self.cr, key)

    def get_bounds(self) -> Rectangle:
        """Return the bounding box."""
        return self._bounds or Rectangle()

    def _update_bounds(self, bounds):
        if bounds:
            if not self._bounds:
                self._bounds = bounds
            else:
                self._bounds += bounds

    def _extents(self, extents_func, line_width=False):
        """Calculate the bounding box for a given drawing operation.

        if ``line_width`` is True, the current line-width is taken into
        account.
        """
        cr = self.cr
        cr.save()
        cr.identity_matrix()
        x0, y0, x1, y1 = extents_func()
        b = Rectangle(x0, y0, x1=x1, y1=y1)
        cr.restore()
        if b and line_width:
            # Do this after the restore(), so we can get the proper width.
            lw = cr.get_line_width() / 2
            d = cr.user_to_device_distance(lw, lw)
            b.expand(d[0] + d[1])
        self._update_bounds(b)
        return b

    def fill(self, b=None):
        """Interceptor for Cairo drawing method."""
        cr = self.cr
        if not b:
            b = self._extents(cr.fill_extents)
        cr.fill()

    def fill_preserve(self, b=None):
        """Interceptor for Cairo drawing method."""
        if not b:
            cr = self.cr
            b = self._extents(cr.fill_extents)

    def stroke(self, b=None):
        """Interceptor for Cairo drawing method."""
        cr = self.cr
        if not b:
            b = self._extents(cr.stroke_extents, line_width=True)
        cr.stroke()

    def stroke_preserve(self, b=None):
        """Interceptor for Cairo drawing method."""
        if not b:
            cr = self.cr
            b = self._extents(cr.stroke_extents, line_width=True)

    def show_text(self, utf8, b=None):
        """Interceptor for Cairo drawing method."""
        cr = self.cr
        if not b:
            x, y = cr.get_current_point()
            e = cr.text_extents(utf8)
            x0, y0 = cr.user_to_device(x + e[0], y + e[1])
            x1, y1 = cr.user_to_device(x + e[0] + e[2], y + e[1] + e[3])
            b = Rectangle(x0, y0, x1=x1, y1=y1)
            self._update_bounds(b)
        cr.show_text(utf8)


class BoundingBoxPainter:
    """This specific case of an ItemPainter is used to calculate the bounding
    boxes (in cairo device coordinates) for the items."""

    draw_all = True

    def __init__(
        self,
        item_painter: ItemPainterType,
    ):
        self.item_painter = item_painter

    def paint_item(self, item: Item, cairo: CairoContext) -> Rectangle:
        bbctx = CairoBoundingBoxContext(cairo)
        self.item_painter.paint_item(item, bbctx)
        # Bounding box is in view (cairo root) coordinates
        bounds = bbctx.get_bounds()

        # Update bounding box with handles.
        i2c = item.matrix_i2c.transform_point
        for h in item.handles():
            cx, cy = cairo.user_to_device(*i2c(*h.pos))
            bounds += (cx - 5, cy - 5, 9, 9)

        bounds.expand(1)
        return bounds

    def paint(
        self, items: Collection[Item], cairo: CairoContext
    ) -> Dict[Item, Rectangle]:
        """Draw the items, return the bounding boxes (in cairo device
        coordinates)."""
        paint_item = self.paint_item
        boxes: Dict[Item, Rectangle] = {item: paint_item(item, cairo) for item in items}
        return boxes

    def bounding_box(self, items: Collection[Item], cairo: CairoContext) -> Rectangle:
        """Get the unified bounding box of the rendered items."""
        boxes = self.paint(items, cairo)
        return functools.reduce(operator.add, boxes.values())



import logging
from functools import singledispatch
from typing import Optional, Sequence

from gi.repository import Gdk

from gaphas.aspect.connector import ConnectionSink, ConnectionSinkType, Connector
from gaphas.connector import Handle
from gaphas.item import Element, Item
from gaphas.types import Pos
from gaphas.view import GtkView

log = logging.getLogger(__name__)


class ItemHandleMove:
    """Move a handle (role is applied to the handle)"""

    GLUE_DISTANCE = 10

    last_x: float
    last_y: float

    def __init__(self, item: Item, handle: Handle, view: GtkView):
        self.item = item
        self.handle = handle
        self.view = view

    def start_move(self, pos: Pos) -> None:
        self.last_x, self.last_y = pos
        model = self.view.model
        assert model
        cinfo = model.connections.get_connection(self.handle)
        if cinfo:
            model.connections.solver.remove_constraint(cinfo.constraint)

    def move(self, pos: Pos) -> None:
        item = self.item
        view = self.view
        assert view.model

        v2i = view.get_matrix_v2i(item)

        x, y = v2i.transform_point(*pos)

        self.handle.pos = (x, y)

        self.glue(pos)

        # do not request matrix update as matrix recalculation will be
        # performed due to item normalization if required
        view.model.request_update(item, matrix=False)

    def stop_move(self, pos: Pos) -> None:
        self.connect(pos)

    def glue(
        self, pos: Pos, distance: int = GLUE_DISTANCE
    ) -> Optional[ConnectionSinkType]:
        """Glue to an item near a specific point.

        Returns a ConnectionSink or None.
        """
        item = self.item
        handle = self.handle
        view = self.view
        model = view.model
        assert model
        connections = model.connections

        if not handle.connectable:
            return None

        connectable = item_at_point(view, pos, distance=distance, exclude=(item,))
        if not connectable:
            return None

        connector = Connector(self.item, self.handle, connections)
        sink = ConnectionSink(connectable)

        if connector.glue(sink):
            return sink

        return None

    def connect(self, pos: Pos) -> None:
        """Connect a handle of a item to connectable item.

        Connectable item is found by `ConnectHandleTool.glue` method.

        :Parameters:
         item
            Connecting item.
         handle
            Handle of connecting item.
         pos
            Position to connect to (or near at least)
        """
        handle = self.handle
        model = self.view.model
        assert model
        connections = model.connections
        connector = Connector(self.item, handle, connections)

        # find connectable item and its port
        sink = self.glue(pos)

        # no new connectable item, then diconnect and exit
        if sink:
            connector.connect(sink)
        else:
            cinfo = connections.get_connection(handle)
            if cinfo:
                connector.disconnect()

        model.request_update(self.item)


HandleMove = singledispatch(ItemHandleMove)


@HandleMove.register(Element)
class ElementHandleMove(ItemHandleMove):
    CURSORS = ("nw-resize", "ne-resize", "se-resize", "sw-resize")

    def __init__(self, item: Item, handle: Handle, view: GtkView):
        super().__init__(item, handle, view)
        self.cursor: Optional[Gdk.Cursor] = None

    def start_move(self, pos: Pos) -> None:
        super().start_move(pos)
        self.set_cursor()

    def stop_move(self, pos: Pos) -> None:
        self.reset_cursor()
        super().stop_move(pos)

    def set_cursor(self) -> None:
        index = self.item.handles().index(self.handle)
        if index < 4:
            display = self.view.get_display()
            cursor = Gdk.Cursor.new_from_name(display, self.CURSORS[index])
            self.cursor = self.view.get_window().get_cursor()
            self.view.get_window().set_cursor(cursor)

    def reset_cursor(self) -> None:
        self.view.get_window().set_cursor(self.cursor)


# Maybe make this an iterator? so extra checks can be done on the item
def item_at_point(
    view: GtkView,
    pos: Pos,
    distance: float = 0.5,
    exclude: Sequence[Item] = (),
) -> Optional[Item]:
    """Return the topmost item located at ``pos`` (x, y).

    Parameters:
        - view: a view
        - pos: Position, a tuple ``(x, y)`` in view coordinates
        - selected: if False returns first non-selected item
    """
    item: Item
    vx, vy = pos
    rect = (vx - distance, vy - distance, distance * 2, distance * 2)
    for item in reversed(list(view.get_items_in_rectangle(rect))):
        if item in exclude:
            continue

        v2i = view.get_matrix_v2i(item)
        ix, iy = v2i.transform_point(vx, vy)
        item_distance = item.point(ix, iy)
        if item_distance is None:
            log.warning("Item distance is None for %s", item)
            continue
        if item_distance < distance:
            return item
    return None



"""Defines aspects for Items.

Aspects form intermediate items between tools and items.
"""

from gaphas.aspect.connector import ConnectionSink, Connector
from gaphas.aspect.handlemove import HandleMove, item_at_point
from gaphas.aspect.move import Move



from typing import Callable, Optional

from gi.repository import Gtk

from gaphas.decorators import g_async
from gaphas.geometry import Rectangle
from gaphas.matrix import Matrix


class Scrolling:
    """Contains Gtk.Adjustment and related code."""

    def __init__(self, scrolling_updated: Callable[[Matrix], None]) -> None:
        self._scrolling_updated = scrolling_updated
        self.hadjustment: Optional[Gtk.Adjustment] = None
        self.vadjustment: Optional[Gtk.Adjustment] = None
        self.hscroll_policy: Optional[Gtk.ScrollablePolicy] = None
        self.vscroll_policy: Optional[Gtk.ScrollablePolicy] = None
        self._hadjustment_handler_id = 0
        self._vadjustment_handler_id = 0

    def get_property(self, prop):
        if prop.name == "hadjustment":
            return self.hadjustment
        elif prop.name == "vadjustment":
            return self.vadjustment
        elif prop.name == "hscroll-policy":
            return self.hscroll_policy
        elif prop.name == "vscroll-policy":
            return self.vscroll_policy
        else:
            raise AttributeError(f"Unknown property {prop.name}")

    def set_property(self, prop, value):
        if prop.name == "hadjustment":
            if value is not None:
                self.hadjustment = value
                self.hadjustment_handler_id = self.hadjustment.connect(
                    "value-changed", self.on_adjustment_changed
                )
                self._scrolling_updated(Matrix())
        elif prop.name == "vadjustment":
            if value is not None:
                self.vadjustment = value
                self.vadjustment_handler_id = self.vadjustment.connect(
                    "value-changed", self.on_adjustment_changed
                )
                self._scrolling_updated(Matrix())
        elif prop.name == "hscroll-policy":
            self.hscroll_policy = value
        elif prop.name == "vscroll-policy":
            self.vscroll_policy = value
        else:
            raise AttributeError(f"Unknown property {prop.name}")

    @g_async(single=True)
    def update_adjustments(self, allocation, bounds):
        aw, ah = allocation.width, allocation.height

        # canvas limits (in view coordinates)
        c = Rectangle(*bounds)

        # view limits
        v = Rectangle(0, 0, aw, ah)

        # union of these limits gives scrollbar limits
        u = c if v in c else c + v
        if not self.hadjustment:
            self._hadjustment = Gtk.Adjustment.new(
                value=v.x,
                lower=u.x,
                upper=u.x1,
                step_increment=aw // 10,
                page_increment=aw,
                page_size=aw,
            )
        else:
            self.hadjustment.set_value(v.x)
            self.hadjustment.set_lower(u.x)
            self.hadjustment.set_upper(u.x1)
            self.hadjustment.set_step_increment(aw // 10)
            self.hadjustment.set_page_increment(aw)
            self.hadjustment.set_page_size(aw)

        if not self.vadjustment:
            self.vadjustment = Gtk.Adjustment.new(
                value=v.y,
                lower=u.y,
                upper=u.y1,
                step_increment=ah // 10,
                page_increment=ah,
                page_size=ah,
            )
        else:
            self.vadjustment.set_value(v.y)
            self.vadjustment.set_lower(u.y)
            self.vadjustment.set_upper(u.y1)
            self.vadjustment.set_step_increment(ah // 10)
            self.vadjustment.set_page_increment(ah)
            self.vadjustment.set_page_size(ah)

    def on_adjustment_changed(self, adj):
        """Change the transformation matrix of the view to reflect the value of
        the x/y adjustment (scrollbar)."""
        value = adj.get_value()
        if value == 0.0:
            return

        # Can not use self._matrix.translate(-value , 0) here, since
        # the translate method effectively does a m * self._matrix, which
        # will result in the translation being multiplied by the orig. matrix
        m = Matrix()
        if adj is self.hadjustment:
            m.translate(-value, 0)
        elif adj is self.vadjustment:
            m.translate(0, -value)

        self._scrolling_updated(m)



from typing import Callable, Collection, Optional, Set

from gaphas.item import Item


class Selection:
    def __init__(self):
        super().__init__()
        self._selected_items: Set[Item] = set()
        self._focused_item: Optional[Item] = None
        self._hovered_item: Optional[Item] = None
        self._handlers: Set[Callable[[], None]] = set()

    def add_handler(self, handler: Callable[[], None]) -> None:
        """Add a callback handler, triggered when a constraint is resolved."""
        self._handlers.add(handler)

    def remove_handler(self, handler: Callable[[], None]) -> None:
        """Remove a previously assigned handler."""
        self._handlers.discard(handler)

    def notify(self) -> None:
        for handler in self._handlers:
            handler()

    def clear(self):
        self._selected_items.clear()
        self._focused_item = None
        self._hovered_item = None
        self.notify()

    @property
    def selected_items(self) -> Collection[Item]:
        return self._selected_items

    def select_items(self, *items: Item) -> None:
        for item in items:
            if item not in self._selected_items:
                self._selected_items.add(item)
                self.notify()

    def unselect_item(self, item: Item) -> None:
        """Unselect an item.

        If it's focused, it will be unfocused as well.
        """
        if item is self._focused_item:
            self._focused_item = None
        if item in self._selected_items:
            self._selected_items.discard(item)
            self.notify()

    def unselect_all(self) -> None:
        """Clearing the selected_item also clears the focused_item."""
        for item in list(self._selected_items):
            self.unselect_item(item)
        self.focused_item = None

    @property
    def focused_item(self) -> Optional[Item]:
        return self._focused_item

    @focused_item.setter
    def focused_item(self, item: Optional[Item]) -> None:
        if item:
            self.select_items(item)

        if item is not self._focused_item:
            self._focused_item = item
            self.notify()

    @property
    def hovered_item(self) -> Optional[Item]:
        return self._hovered_item

    @hovered_item.setter
    def hovered_item(self, item: Optional[Item]) -> None:
        if item is not self._hovered_item:
            self._hovered_item = item
            self.notify()



from gi.repository import Gtk

from gaphas.view import GtkView


def scroll_tool(view: GtkView, speed: int = 5) -> Gtk.EventControllerScroll:
    """Scroll tool recognized 2 finger scroll gestures."""
    ctrl = Gtk.EventControllerScroll.new(
        view,
        Gtk.EventControllerScrollFlags.BOTH_AXES,
    )
    ctrl.connect("scroll", on_scroll, speed)
    return ctrl


def on_scroll(controller, dx, dy, speed):
    view = controller.get_widget()
    m = view.matrix
    m.translate(-dx * speed, -dy * speed)
    view.request_update((), view.model.get_all_items())



"""
Quadtree
========

A quadtree is a tree data structure in which each internal node has up
to four children. Quadtrees are most often used to partition a two
dimensional space by recursively subdividing it into four quadrants or
regions. The regions may be square or rectangular, or may have
arbitrary shapes. This data structure was named a quadtree by Raphael
Finkel and J.L. Bentley in 1974. A similar partitioning is also known
as a Q-tree. All forms of Quadtrees share some common features:

* They decompose space into adaptable cells.
* Each cell (or bucket) has a maximum capacity.
  When maximum capacity is reached, the bucket splits.
* The tree directory follows the spatial decomposition of the Quadtree.

(From Wikipedia, the free encyclopedia)
"""
from __future__ import annotations

import operator
from typing import (
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
)

from gaphas.geometry import rectangle_clip, rectangle_contains, rectangle_intersects

Bounds = Tuple[float, float, float, float]  # x, y, width, height

T = TypeVar("T")
D = TypeVar("D")


class Quadtree(Generic[T, D]):
    """The Quad-tree.

    Rectangles use the same scheme throughout Gaphas: (x, y, width, height).

    >>> qtree = Quadtree((0, 0, 100, 100))
    >>> for i in range(20):
    ...     qtree.add(f"{i}", ((i * 4) % 90, (i * 10) % 90, 10, 10))
    >>> len(qtree)
    20
    >>> qtree.dump() # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
     <....QuadtreeBucket object at 0x...> (0, 0, 100, 100)
       11 (44, 20, 10, 10)
       12 (48, 30, 10, 10)
       <....QuadtreeBucket object at 0x...> (0, 0, 50.0, 50.0)
         0 (0, 0, 10, 10)
         1 (4, 10, 10, 10)
         2 (8, 20, 10, 10)
         3 (12, 30, 10, 10)
         4 (16, 40, 10, 10)
         9 (36, 0, 10, 10)
         10 (40, 10, 10, 10)
       <....QuadtreeBucket object at 0x...> (50.0, 0, 50.0, 50.0)
         13 (52, 40, 10, 10)
         18 (72, 0, 10, 10)
         19 (76, 10, 10, 10)
       <....QuadtreeBucket object at 0x...> (0, 50.0, 50.0, 50.0)
         5 (20, 50, 10, 10)
         6 (24, 60, 10, 10)
         7 (28, 70, 10, 10)
         8 (32, 80, 10, 10)
       <....QuadtreeBucket object at 0x...> (50.0, 50.0, 50.0, 50.0)
         14 (56, 50, 10, 10)
         15 (60, 60, 10, 10)
         16 (64, 70, 10, 10)
         17 (68, 80, 10, 10)

    Find all items in the tree:

    >>> sorted(qtree.find_inside((0, 0, 100, 100)))
    ['0', '1', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '2', '3', '4', '5', '6', '7', '8', '9']

    Or just the items in a section of the tree:

    >>> sorted(qtree.find_inside((40, 40, 40, 40)))
    ['13', '14', '15', '16']
    """

    def __init__(self, bounds: Bounds = (0, 0, 0, 0), capacity: int = 10) -> None:
        """Create a new Quadtree instance.

        Bounds is the boundaries of the quadtree. this is fixed and do not
        change depending on the contents.

        Capacity defines the number of elements in one tree bucket (default: 10)
        """
        self._capacity = capacity
        self._bucket: QuadtreeBucket[T] = QuadtreeBucket(bounds, capacity)

        # Easy lookup item->(bounds, data, clipped bounds)
        self._ids: Dict[T, Tuple[Bounds, D, Optional[Bounds]]] = {}

    @property
    def bounds(self) -> Bounds:
        return self._bucket.bounds

    def resize(self, bounds: Bounds) -> None:
        """Resize the tree.

        The tree structure is rebuild.
        """
        self._bucket = QuadtreeBucket(bounds, self._capacity)
        self.rebuild()

    @property
    def soft_bounds(self) -> Bounds:
        """Calculate the size of all items in the tree. This size may be beyond
        the limits of the tree itself.

        Returns a tuple (x, y, width, height).

        >>> qtree = Quadtree()
        >>> qtree.add('1', (10, 20, 30, 40))
        >>> qtree.add('2', (20, 30, 40, 10))
        >>> qtree.soft_bounds
        (10, 20, 50, 40)

        Quadtree's bounding box is not adjusted:

        >>> qtree.bounds
        (0, 0, 0, 0)
        """
        x_y_w_h = list(zip(*[d[0] for d in self._ids.values()]))
        if not x_y_w_h:
            return 0, 0, 0, 0
        x0 = min(x_y_w_h[0])
        y0 = min(x_y_w_h[1])
        add = operator.add
        x1 = max(map(add, x_y_w_h[0], x_y_w_h[2]))
        y1 = max(map(add, x_y_w_h[1], x_y_w_h[3]))
        return x0, y0, x1 - x0, y1 - y0

    def add(self, item: T, bounds: Bounds, data: D) -> None:
        """Add an item to the tree.

        If an item already exists, its bounds are updated and the item
        is moved to the right bucket. Data can be used to add some extra
        info to the item
        """
        # Clip item bounds to fit in top-level bucket
        # Keep original bounds in _ids, for reference
        clipped_bounds = rectangle_clip(bounds, self._bucket.bounds)

        if item in self._ids:
            old_clip = self._ids[item][2]
            if old_clip:
                bucket = self._bucket.find_bucket(old_clip)
                assert item in bucket.items
                # Fast lane, if item moved just a little it may still reside
                # in the same bucket. We do not need to search from top-level.
                if (
                    bucket
                    and clipped_bounds
                    and rectangle_contains(clipped_bounds, bucket.bounds)
                ):
                    bucket.update(item, clipped_bounds)
                    self._ids[item] = (bounds, data, clipped_bounds)
                    return
                elif bucket:
                    bucket.remove(item)

        if clipped_bounds:
            self._bucket.find_bucket(clipped_bounds).add(item, clipped_bounds)
        self._ids[item] = (bounds, data, clipped_bounds)

    def remove(self, item: T) -> None:
        """Remove an item from the tree."""
        bounds, data, clipped_bounds = self._ids[item]
        del self._ids[item]
        if clipped_bounds:
            self._bucket.find_bucket(clipped_bounds).remove(item)

    def clear(self):
        """Remove all items from the tree."""
        self._bucket.clear()
        self._ids.clear()

    def rebuild(self):
        """Rebuild the tree structure."""
        # Clean bucket and items:
        self._bucket.clear()

        for item, (bounds, data, _) in dict(self._ids).items():
            clipped_bounds = rectangle_clip(bounds, self._bucket.bounds)
            if clipped_bounds:
                self._bucket.find_bucket(clipped_bounds).add(item, clipped_bounds)
            self._ids[item] = (bounds, data, clipped_bounds)

    def get_bounds(self, item: T) -> Bounds:
        """Return the bounding box for the given item."""
        return self._ids[item][0]

    def get_data(self, item: T) -> D:
        """Return the data for the given item, None if no data was provided."""
        return self._ids[item][1]

    def get_clipped_bounds(self, item: T) -> Optional[Bounds]:
        """Return the bounding box for the given item.

        The bounding box is clipped on the boundaries of the tree
        (provided on construction or with resize()).
        """
        return self._ids[item][2]

    def find_inside(self, rect: Bounds) -> Set[T]:
        """Find all items in the given rectangle (x, y, with, height).

        Returns a set.
        """
        return set(self._bucket.find(rect, method=rectangle_contains))

    def find_intersect(self, rect: Bounds) -> Set[T]:
        """Find all items that intersect with the given rectangle (x, y, width,
        height).

        Returns a set.
        """
        return set(self._bucket.find(rect, method=rectangle_intersects))

    def __len__(self):
        """Return number of items in tree."""
        return len(self._ids)

    def __contains__(self, item: T) -> bool:
        """Check if an item is in tree."""
        return item in self._ids

    def dump(self) -> None:
        """Print structure to stdout."""
        self._bucket.dump()


class QuadtreeBucket(Generic[T]):
    """A node in a Quadtree structure."""

    def __init__(self, bounds: Bounds, capacity: int):
        """Set bounding box for the node as (x, y, width, height)."""
        self.bounds = bounds
        self.capacity = capacity

        self.items: Dict[T, Bounds] = {}
        self._buckets: List[QuadtreeBucket[T]] = []

    def add(self, item: T, bounds: Bounds) -> None:
        """Add an item to the quadtree.

        The bucket is split when necessary. Items are otherwise added to
        this bucket, not some sub-bucket.
        """
        assert rectangle_contains(bounds, self.bounds)
        # create new subnodes if threshold is reached
        if not self._buckets and len(self.items) >= self.capacity:
            x, y, w, h = self.bounds
            rw, rh = w / 2.0, h / 2.0
            cx, cy = x + rw, y + rh
            self._buckets = [
                QuadtreeBucket((x, y, rw, rh), self.capacity),
                QuadtreeBucket((cx, y, rw, rh), self.capacity),
                QuadtreeBucket((x, cy, rw, rh), self.capacity),
                QuadtreeBucket((cx, cy, rw, rh), self.capacity),
            ]
            # Add items to subnodes
            items = list(self.items.items())
            self.items.clear()
            for i, b in items:
                self.find_bucket(b).add(i, b)
            self.find_bucket(bounds).add(item, bounds)
        else:
            self.items[item] = bounds

    def remove(self, item: T) -> None:
        """Remove an item from the quadtree bucket.

        The item should be contained by *this* bucket (not a sub-
        bucket).
        """
        del self.items[item]

    def update(self, item: T, new_bounds: Bounds) -> None:
        """Update the position of an item within the current bucket.

        The item should live in the current bucket, but may be placed in
        a sub-bucket.
        """
        assert item in self.items
        self.remove(item)
        self.find_bucket(new_bounds).add(item, new_bounds)

    def find_bucket(self, bounds: Bounds) -> QuadtreeBucket:
        """Find the bucket that holds a bounding box.

        This method should be used to find a bucket that fits, before
        add() or remove() is called.
        """
        if not self._buckets:
            return self
        sx, sy, sw, sh = self.bounds
        cx, cy = sx + sw / 2.0, sy + sh / 2.0
        x, y, w, h = bounds
        index = 0
        if x >= cx:
            index += 1
        elif x + w > cx:
            return self

        if y >= cy:
            index += 2
        elif y + h > cy:
            return self
        return self._buckets[index].find_bucket(bounds)

    def find(
        self, rect: Bounds, method: Callable[[Bounds, Bounds], bool]
    ) -> Iterable[T]:
        """Find all items in the given rectangle (x, y, with, height). Method
        can be either the contains or intersects function.

        Returns an iterator.
        """
        if rectangle_intersects(rect, self.bounds):
            for item, bounds in list(self.items.items()):
                if method(bounds, rect):
                    yield item
            for bucket in self._buckets:
                yield from bucket.find(rect, method=method)

    def clear(self):
        """Clear the bucket, including sub-buckets."""
        del self._buckets[:]
        self.items.clear()

    def dump(self, indent=""):
        print(indent, self, self.bounds)
        indent += "   "
        for item, bounds in sorted(self.items.items(), key=lambda items: items[1]):
            print(indent, item, bounds)
        for bucket in self._buckets:
            bucket.dump(indent)



"""Constraint solver allows to define constraint between two or more different
variables and keep this constraint always true when one or more of the
constrained variables change. For example, one may want to keep two variables
always equal.

Variables change and at some point of time we want to make all
constraints valid again. This process is called solving constraints.

Gaphas' solver allows to define constraints between Variable
instances.  Constraint classes are defined in `gaphas.constraint`
module.

How It Works
------------
Every constraint contains list of variables and has to be registered
in solver object. Variables change (`Variable.dirty()`,
`Solver.request_resolve()` methods) and their constraints are marked
by solver as dirty. To solve constraints, solver loops through dirty
constraints and asks constraint for
a variable (called weakest variable), which

- has the lowest strength
- or if there are many variables with the same, lowest strength value
  return first unchanged variable with lowest strength
- or if there is no unchanged, then return the first changed with the
  lowest strength

(weakest variable invariants defined above)

Having weakest variable (`constraint.Constraint.weakest()` method)
every constraint is being asked to solve itself
(`constraint.Constraint.solve_for()` method) changing appropriate
variables to make the constraint valid again.
"""
import functools
from typing import Callable, Collection, List, Optional, Set

from gaphas.solver.constraint import Constraint, ContainsConstraints


class Solver:
    """Solve constraints.

    A constraint should have accompanying variables.
    """

    def __init__(self) -> None:
        # a dict of constraint -> name/variable mappings
        self._constraints: Set[Constraint] = set()
        self._marked_cons: List[Constraint] = []
        self._solving = False
        self._handlers: Set[Callable[[Constraint], None]] = set()

    def add_handler(self, handler: Callable[[Constraint], None]) -> None:
        """Add a callback handler, triggered when a constraint is resolved."""
        self._handlers.add(handler)

    def remove_handler(self, handler: Callable[[Constraint], None]) -> None:
        """Remove a previously assigned handler."""
        self._handlers.discard(handler)

    def _notify(self, constraint: Constraint) -> None:
        my_constraint = self._find_containing_constraint(constraint)
        for handler in self._handlers:
            handler(my_constraint)

    @functools.lru_cache()
    def _find_containing_constraint(self, constraint: Constraint) -> Constraint:
        return find_containing_constraint(constraint, self._constraints) or constraint

    @property
    def constraints(self) -> Collection[Constraint]:
        return self._constraints

    def add_constraint(self, constraint: Constraint) -> Constraint:
        """Add a constraint. The actual constraint is returned, so the
        constraint can be removed later on.

        Example:

        >>> from gaphas.constraint import EquationConstraint
        >>> s = Solver()
        >>> a, b = Variable(), Variable(2.0)
        >>> s.add_constraint(EquationConstraint(lambda a, b: a -b, a=a, b=b))
        EquationConstraint(<lambda>, a=Variable(0, 20), b=Variable(2, 20))
        >>> len(s._constraints)
        1
        >>> a.value
        0.0
        >>> b.value
        2.0
        >>> len(s._constraints)
        1
        """
        assert constraint, f"No constraint ({constraint})"
        self._constraints.add(constraint)
        self._marked_cons.append(constraint)
        constraint.add_handler(self.request_resolve_constraint)
        return constraint

    def remove_constraint(self, constraint: Constraint) -> None:
        """Remove a constraint from the solver.

        >>> from gaphas.constraint import EquationConstraint
        >>> s = Solver()
        >>> a, b = Variable(), Variable(2.0)
        >>> c = s.add_constraint(EquationConstraint(lambda a, b: a -b, a=a, b=b))
        >>> c
        EquationConstraint(<lambda>, a=Variable(0, 20), b=Variable(2, 20))
        >>> s.remove_constraint(c)
        >>> s._marked_cons
        []
        >>> s._constraints
        set()

        Removing a constraint twice has no effect:

        >>> s.remove_constraint(c)
        """
        assert constraint, f"No constraint ({constraint})"
        constraint.remove_handler(self.request_resolve_constraint)
        self._constraints.discard(constraint)
        while constraint in self._marked_cons:
            self._marked_cons.remove(constraint)

    def request_resolve_constraint(self, c: Constraint) -> None:
        """Request resolving a constraint."""
        if not self._solving:
            if c in self._marked_cons:
                self._marked_cons.remove(c)
            self._marked_cons.append(c)
        else:
            self._marked_cons.append(c)
            if self._marked_cons.count(c) > 100:
                raise JuggleError(
                    f"Variable juggling detected, constraint {c} resolved {self._marked_cons.count(c)} times out of {len(self._marked_cons)}"
                )

    def solve(self) -> None:
        """
        Example:

        >>> from gaphas.constraint import EquationConstraint
        >>> a, b, c = Variable(1.0), Variable(2.0), Variable(3.0)
        >>> s = Solver()
        >>> s.add_constraint(EquationConstraint(lambda a,b: a+b, a=a, b=b))
        EquationConstraint(<lambda>, a=Variable(1, 20), b=Variable(2, 20))
        >>> a.value = 5.0
        >>> s.solve()
        >>> len(s._marked_cons)
        0
        >>> b._value
        -5.0
        >>> s.add_constraint(EquationConstraint(lambda a,b: a+b, a=b, b=c))
        EquationConstraint(<lambda>, a=Variable(-5, 20), b=Variable(3, 20))
        >>> len(s._constraints)
        2
        >>> len(s._marked_cons)
        1
        >>> b._value
        -5.0
        >>> s.solve()
        >>> b._value
        -3.0
        >>> a.value = 10
        >>> s.solve()
        >>> c._value
        10.0
        """
        marked_cons = self._marked_cons
        notify = self._notify
        try:
            self._solving = True

            # Solve each constraint. Using a counter makes it
            # possible to also solve constraints that are marked as
            # a result of other variabled being solved.
            n = 0
            while n < len(marked_cons):
                c = marked_cons[n]
                c.solve()
                notify(c)
                n += 1

            self._marked_cons = []
        finally:
            self._solving = False


def find_containing_constraint(
    constraint: Constraint, constraints: Collection[Constraint]
) -> Optional[Constraint]:
    if constraint in constraints:
        return constraint

    for cs in constraints:
        if isinstance(cs, ContainsConstraints) and find_containing_constraint(
            constraint, cs.constraints
        ):
            return find_containing_constraint(cs, constraints)
    return None


class JuggleError(AssertionError):
    """Variable juggling exception.

    Raised when constraint's variables are marking each other dirty
    forever.
    """


__test__ = {
    "Solver.add_constraint": Solver.add_constraint,
    "Solver.remove_constraint": Solver.remove_constraint,
}



"""This module contains everything to display a model on a screen."""
from __future__ import annotations

from typing import Collection, Iterable, Optional, Set, Tuple

import cairo
from gi.repository import Gdk, GLib, GObject, Gtk

from gaphas.canvas import instant_cairo_context
from gaphas.decorators import g_async
from gaphas.geometry import Rect, Rectangle
from gaphas.item import Item
from gaphas.matrix import Matrix
from gaphas.painter import BoundingBoxPainter, DefaultPainter, ItemPainter, Painter
from gaphas.quadtree import Quadtree, QuadtreeBucket
from gaphas.view.model import Model
from gaphas.view.scrolling import Scrolling
from gaphas.view.selection import Selection

# Handy debug flag for drawing bounding boxes around the items.
DEBUG_DRAW_BOUNDING_BOX = False
DEBUG_DRAW_QUADTREE = False

# The default cursor (use in case of a cursor reset)
DEFAULT_CURSOR = Gdk.CursorType.LEFT_PTR


class GtkView(Gtk.DrawingArea, Gtk.Scrollable):
    """GTK+ widget for rendering a gaphas.view.model.Model to a screen.  The
    view uses Tools to handle events and Painters to draw. Both are
    configurable.

    The widget already contains adjustment objects (`hadjustment`,
    `vadjustment`) to be used for scrollbars.

    This view registers itself on the model, so it will receive
    update events.
    """

    # Just defined a name to make GTK register this class.
    __gtype_name__ = "GaphasView"

    # properties required by Gtk.Scrollable
    __gproperties__ = {
        "hscroll-policy": (
            Gtk.ScrollablePolicy,
            "hscroll-policy",
            "hscroll-policy",
            Gtk.ScrollablePolicy.MINIMUM,
            GObject.ParamFlags.READWRITE,
        ),
        "hadjustment": (
            Gtk.Adjustment,
            "hadjustment",
            "hadjustment",
            GObject.ParamFlags.READWRITE,
        ),
        "vscroll-policy": (
            Gtk.ScrollablePolicy,
            "vscroll-policy",
            "vscroll-policy",
            Gtk.ScrollablePolicy.MINIMUM,
            GObject.ParamFlags.READWRITE,
        ),
        "vadjustment": (
            Gtk.Adjustment,
            "vadjustment",
            "vadjustment",
            GObject.ParamFlags.READWRITE,
        ),
    }

    def __init__(self, model: Optional[Model] = None, selection: Selection = None):
        """Create a new view.

        Args:
            model (Model): optional model to be set on construction time.
            selection (Selection): optional selection object, in case the default
                selection object (hover/select/focus) is not enough.
        """
        Gtk.DrawingArea.__init__(self)

        self._dirty_items: Set[Item] = set()
        self._dirty_matrix_items: Set[Item] = set()

        self._back_buffer: Optional[cairo.Surface] = None
        self._back_buffer_needs_resizing = True

        self._controllers: Set[Gtk.EventController] = set()

        self.set_can_focus(True)
        self.add_events(
            Gdk.EventMask.BUTTON_PRESS_MASK
            | Gdk.EventMask.BUTTON_RELEASE_MASK
            | Gdk.EventMask.POINTER_MOTION_MASK
            | Gdk.EventMask.KEY_PRESS_MASK
            | Gdk.EventMask.KEY_RELEASE_MASK
            | Gdk.EventMask.SCROLL_MASK
            | Gdk.EventMask.STRUCTURE_MASK
        )

        def alignment_updated(matrix: Matrix) -> None:
            assert self._model
            self._matrix *= matrix  # type: ignore[operator]

            # Force recalculation of the bounding boxes:
            self.request_update((), self._model.get_all_items())

        self._scrolling = Scrolling(alignment_updated)

        self._selection = selection or Selection()

        self._matrix = Matrix()
        self._painter: Painter = DefaultPainter(self)
        self._bounding_box_painter: Painter = BoundingBoxPainter(
            ItemPainter(self._selection)
        )

        self._qtree: Quadtree[Item, Tuple[float, float, float, float]] = Quadtree()

        self._model: Optional[Model] = None
        if model:
            self.model = model

        self._selection.add_handler(self.queue_redraw)

    def do_get_property(self, prop: str) -> object:
        return self._scrolling.get_property(prop)

    def do_set_property(self, prop: str, value: object) -> None:
        self._scrolling.set_property(prop, value)

    @property
    def matrix(self) -> Matrix:
        """Model root to view transformation matrix."""
        return self._matrix

    def get_matrix_i2v(self, item: Item) -> Matrix:
        """Get Item to View matrix for ``item``."""
        return item.matrix_i2c.multiply(self._matrix)

    def get_matrix_v2i(self, item: Item) -> Matrix:
        """Get View to Item matrix for ``item``."""
        m = self.get_matrix_i2v(item)
        m.invert()
        return m

    @property
    def model(self) -> Optional[Model]:
        """The model."""
        return self._model

    @model.setter
    def model(self, model: Optional[Model]) -> None:
        if self._model:
            self._model.unregister_view(self)
            self._selection.clear()
            self._qtree.clear()

        self._model = model

        if self._model:
            self._model.register_view(self)
            self.request_update(self._model.get_all_items())

    @property
    def painter(self) -> Painter:
        """Painter for drawing the view."""
        return self._painter

    @painter.setter
    def painter(self, painter: Painter) -> None:
        self._painter = painter

    @property
    def bounding_box_painter(self) -> Painter:
        """Special painter for calculating item bounding boxes."""
        return self._bounding_box_painter

    @bounding_box_painter.setter
    def bounding_box_painter(self, painter: Painter) -> None:
        self._bounding_box_painter = painter

    @property
    def selection(self) -> Selection:
        """Selected, focused and hovered items."""
        return self._selection

    @selection.setter
    def selection(self, selection: Selection) -> None:
        """Selected, focused and hovered items."""
        self._selection = selection
        self.queue_redraw()

    @property
    def bounding_box(self) -> Rectangle:
        """The bounding box of the complete view, relative to the view port."""
        return Rectangle(*self._qtree.soft_bounds)

    @property
    def hadjustment(self) -> Gtk.Adjustment:
        """Gtk adjustment object for use with a scrollbar."""
        return self._scrolling.hadjustment

    @property
    def vadjustment(self) -> Gtk.Adjustment:
        """Gtk adjustment object for use with a scrollbar."""
        return self._scrolling.vadjustment

    def add_controller(self, *controllers: Gtk.EventController) -> None:
        """Add a controller.

        A convenience method, so you have a place to store the event
        controllers. Events controllers are linked to a widget (in GTK3)
        on creation time, so calling this method is not necessary.
        """
        self._controllers.update(controllers)

    def remove_controller(self, controller: Gtk.EventController) -> bool:
        """Remove a controller.

        The event controller's propagation phase is set to
        `Gtk.PropagationPhase.NONE` to ensure it's not invoked
        anymore.

        NB. The controller is only really removed from the widget when it's destroyed!
            This is a Gtk3 limitation.
        """
        if controller in self._controllers:
            controller.set_propagation_phase(Gtk.PropagationPhase.NONE)
            self._controllers.discard(controller)
            return True
        return False

    def remove_all_controllers(self) -> None:
        """Remove all registered controllers."""
        for controller in set(self._controllers):
            self.remove_controller(controller)

    def zoom(self, factor: float) -> None:
        """Zoom in/out by factor ``factor``."""
        assert self._model
        self.matrix.scale(factor, factor)
        self.request_update((), self._model.get_all_items())

    def get_items_in_rectangle(
        self, rect: Rect, contain: bool = False
    ) -> Iterable[Item]:
        """Return the items in the rectangle 'rect'.

        Items are automatically sorted in model's processing order.
        """
        assert self._model
        items = (
            self._qtree.find_inside(rect)
            if contain
            else self._qtree.find_intersect(rect)
        )
        return self._model.sort(items)

    def get_item_bounding_box(self, item: Item) -> Rectangle:
        """Get the bounding box for the item, in view coordinates."""
        return Rectangle(*self._qtree.get_bounds(item))

    def queue_redraw(self) -> None:
        """Redraw the entire view."""
        self.update_back_buffer()

    def request_update(
        self,
        items: Iterable[Item],
        matrix_only_items: Iterable[Item] = (),
        removed_items: Iterable[Item] = (),
    ) -> None:
        """Request update for items.

        Items will get a full update treatment, while
        ``matrix_only_items`` will only have their bounding box
        recalculated.
        """
        if items:
            self._dirty_items.update(items)
        if matrix_only_items:
            self._dirty_matrix_items.update(matrix_only_items)

        # Remove removed items:
        if removed_items:
            selection = self._selection
            self._dirty_matrix_items.difference_update(removed_items)
            self._dirty_items.difference_update(removed_items)

            for item in removed_items:
                self._qtree.remove(item)
                selection.unselect_item(item)

        self.update()

    @g_async(single=True)
    def update(self) -> None:
        """Update view status according to the items updated in the model."""
        model = self._model
        if not model:
            return

        try:
            dirty_items = self._dirty_items
            dirty_matrix_items = self.all_dirty_matrix_items()
            dirty_items.update(self.update_qtree(dirty_items, dirty_matrix_items))

            model.update_now(dirty_items, dirty_matrix_items)

            self.update_bounding_box(dirty_items)
            self._scrolling.update_adjustments(
                self.get_allocation(), self._qtree.soft_bounds
            )
            self.update_back_buffer()
        finally:
            self._dirty_items.clear()
            self._dirty_matrix_items.clear()

    def all_dirty_matrix_items(self) -> Set[Item]:
        """Recalculate matrices of the items. Items' children matrices are
        recalculated, too.

        Return items, which matrices were recalculated.
        """
        model = self._model
        if not model:
            return set()

        def update_matrices(items: Iterable[Item]) -> Iterable[Item]:
            assert model
            for item in items:
                parent = model.get_parent(item)
                if parent is not None and parent in items:
                    # item's matrix will be updated thanks to parent's matrix update
                    continue

                yield item

                yield from update_matrices(set(model.get_children(item)))

        return set(update_matrices(self._dirty_matrix_items))

    def update_qtree(
        self, dirty_items: Collection[Item], dirty_matrix_items: Iterable[Item]
    ) -> Iterable[Item]:
        for i in dirty_matrix_items:
            if i not in self._qtree:
                yield i
            elif i not in dirty_items:
                # Only matrix has changed, so calculate new bounding box
                # based on quadtree data (= bb in item coordinates).
                bounds = self._qtree.get_data(i)
                i2v = self.get_matrix_i2v(i)
                x, y = i2v.transform_point(bounds[0], bounds[1])
                w, h = i2v.transform_distance(bounds[2], bounds[3])
                vbounds = Rectangle(x, y, w, h)
                self._qtree.add(i, vbounds.tuple(), bounds)

    def update_bounding_box(self, items: Collection[Item]) -> None:
        """Update the bounding boxes of the model items for this view, in model
        coordinates."""
        cr = (
            cairo.Context(self._back_buffer)
            if self._back_buffer
            else instant_cairo_context()
        )

        cr.set_matrix(self.matrix.to_cairo())
        cr.save()
        cr.rectangle(0, 0, 0, 0)
        cr.clip()
        try:
            painter = self._bounding_box_painter
            if items is None:
                items = list(self.model.get_all_items())

            items_and_bounds = painter.paint(items, cr)
            assert items_and_bounds is not None
            for item, bounds in items_and_bounds.items():
                v2i = self.get_matrix_v2i(item)
                ix, iy = v2i.transform_point(bounds.x, bounds.y)
                iw, ih = v2i.transform_distance(bounds.width, bounds.height)
                self._qtree.add(item=item, bounds=bounds.tuple(), data=(ix, iy, iw, ih))
        finally:
            cr.restore()

    @g_async(single=True, priority=GLib.PRIORITY_HIGH_IDLE)
    def update_back_buffer(self) -> None:
        if self.model and self.get_window():
            if not self._back_buffer or self._back_buffer_needs_resizing:
                allocation = self.get_allocation()
                self._back_buffer = self.get_window().create_similar_surface(
                    cairo.Content.COLOR_ALPHA, allocation.width, allocation.height
                )
                self._back_buffer_needs_resizing = False

            assert self._back_buffer
            allocation = self.get_allocation()
            cr = cairo.Context(self._back_buffer)

            cr.save()
            cr.set_operator(cairo.OPERATOR_CLEAR)
            cr.paint()
            cr.restore()

            Gtk.render_background(
                self.get_style_context(), cr, 0, 0, allocation.width, allocation.height
            )

            items = self.get_items_in_rectangle(
                (0, 0, allocation.width, allocation.height)
            )

            cr.set_matrix(self.matrix.to_cairo())
            cr.save()
            self.painter.paint(list(items), cr)
            cr.restore()

            if DEBUG_DRAW_BOUNDING_BOX:
                for item in self.get_items_in_rectangle(
                    (0, 0, allocation.width, allocation.height)
                ):
                    try:
                        b = self.get_item_bounding_box(item)
                    except KeyError:
                        pass  # No bounding box right now..
                    else:
                        cr.save()
                        cr.identity_matrix()
                        cr.set_source_rgb(0.8, 0, 0)
                        cr.set_line_width(1.0)
                        cr.rectangle(*b)
                        cr.stroke()
                        cr.restore()

            if DEBUG_DRAW_QUADTREE:

                def draw_qtree_bucket(bucket: QuadtreeBucket) -> None:
                    cr.rectangle(*bucket.bounds)
                    cr.stroke()
                    for b in bucket._buckets:
                        draw_qtree_bucket(b)

                cr.set_source_rgb(0, 0, 0.8)
                cr.set_line_width(1.0)
                draw_qtree_bucket(self._qtree._bucket)

            self.get_window().invalidate_rect(allocation, True)

    def do_realize(self) -> None:
        Gtk.DrawingArea.do_realize(self)

        if self._model:
            # Ensure updates are propagated
            self._model.register_view(self)
            self.request_update(self._model.get_all_items())

    def do_unrealize(self) -> None:
        if self._model:
            self._model.unregister_view(self)

        self._qtree.clear()

        self._dirty_items.clear()
        self._dirty_matrix_items.clear()

        Gtk.DrawingArea.do_unrealize(self)

    def do_configure_event(self, event: Gdk.EventConfigure) -> bool:
        allocation = self.get_allocation()
        self._scrolling.update_adjustments(allocation, self._qtree.soft_bounds)
        self._qtree.resize((0, 0, allocation.width, allocation.height))
        if self.get_window():
            self._back_buffer_needs_resizing = True
            self.update_back_buffer()
        else:
            self._back_buffer = None

        return False

    def do_draw(self, cr: cairo.Context) -> bool:
        if not self._model:
            return False

        if not self._back_buffer:
            return False

        cr.set_source_surface(self._back_buffer, 0, 0)
        cr.paint()

        return False



from __future__ import annotations

from typing import Callable, Collection, Set

from typing_extensions import Protocol, runtime_checkable

from gaphas.solver.variable import Variable


@runtime_checkable
class Constraint(Protocol):
    def add_handler(self, handler: Callable[[Constraint], None]) -> None:
        ...

    def remove_handler(self, handler: Callable[[Constraint], None]) -> None:
        ...

    def solve(self) -> None:
        ...


@runtime_checkable
class ContainsConstraints(Protocol):
    @property
    def constraints(self) -> Collection[Constraint]:
        ...


class BaseConstraint:
    """Constraint base class.

    - variables - list of all variables
    - weakest   - list of weakest variables
    """

    def __init__(self, *variables):
        """Create new constraint, register all variables, and find weakest
        variables.

        Any value can be added. It is assumed to be a variable if it has
        a 'strength' attribute.
        """
        self._variables = [v for v in variables if hasattr(v, "strength")]

        strength = min(v.strength for v in self._variables)
        # manage weakest based on id, so variables are uniquely identifyable
        self._weakest = [(id(v), v) for v in self._variables if v.strength == strength]
        self._handlers: Set[Callable[[Constraint], None]] = set()

    def variables(self):
        """Return an iterator which iterates over the variables that are held
        by this constraint."""
        return self._variables

    def add_handler(self, handler: Callable[[Constraint], None]) -> None:
        if not self._handlers:
            for v in self._variables:
                v.add_handler(self._propagate)
        self._handlers.add(handler)

    def remove_handler(self, handler: Callable[[Constraint], None]) -> None:
        self._handlers.discard(handler)
        if not self._handlers:
            for v in self._variables:
                v.remove_handler(self._propagate)

    def notify(self):
        for handler in self._handlers:
            handler(self)

    def _propagate(self, variable, _old):
        self.mark_dirty(variable)
        self.notify()

    def weakest(self):
        """Return the weakest variable.

        The weakest variable should be always as first element of
        Constraint._weakest list.
        """
        return self._weakest[0][1]

    def mark_dirty(self, var: Variable) -> None:
        """Mark variable v dirty and if possible move it to the end of
        Constraint.weakest list to maintain weakest variable invariants (see
        gaphas.solver module documentation)."""

        if isinstance(var, Variable):
            key = (id(var), var)
            try:
                self._weakest.remove(key)
            except ValueError:
                return
            else:
                self._weakest.append(key)
                return

    def solve(self):
        """Solve the constraint.

        This is done by determining the weakest variable and calling
        solve_for() for that variable. The weakest variable is always in
        the set of variables with the weakest strength. The least
        recently changed variable is considered the weakest.
        """
        wvar = self.weakest()
        self.solve_for(wvar)

    def solve_for(self, var):
        """Solve the constraint for a given variable.

        The variable itself is updated.
        """
        raise NotImplementedError


class MultiConstraint:
    """A constraint constaining constraints."""

    def __init__(self, *constraints: Constraint):
        self._constraints = constraints

    @property
    def constraints(self) -> Collection[Constraint]:
        return self._constraints

    def add_handler(self, handler: Callable[[Constraint], None]) -> None:
        for c in self._constraints:
            c.add_handler(handler)

    def remove_handler(self, handler: Callable[[Constraint], None]) -> None:
        for c in self._constraints:
            c.remove_handler(handler)

    def solve(self):
        for c in self._constraints:
            c.solve()



import logging
from typing import Optional, Tuple, Union

from gi.repository import Gdk, Gtk
from typing_extensions import Protocol

from gaphas.aspect import HandleMove, Move, item_at_point
from gaphas.canvas import ancestors
from gaphas.connector import Handle
from gaphas.geometry import distance_point_point_fast
from gaphas.item import Item
from gaphas.types import Pos
from gaphas.view import GtkView

log = logging.getLogger(__name__)


class MoveType(Protocol):
    def __init__(self, item: Item, view: GtkView):
        ...

    def start_move(self, pos: Pos) -> None:
        ...

    def move(self, pos: Pos) -> None:
        ...

    def stop_move(self, pos: Pos) -> None:
        ...


def item_tool(view: GtkView) -> Gtk.GestureDrag:
    """Handle item movement and movement of handles."""
    gesture = Gtk.GestureDrag.new(view)
    drag_state = DragState()
    gesture.connect("drag-begin", on_drag_begin, drag_state)
    gesture.connect("drag-update", on_drag_update, drag_state)
    gesture.connect("drag-end", on_drag_end, drag_state)
    return gesture


class DragState:
    def __init__(self):
        self.moving = set()


def on_drag_begin(gesture, start_x, start_y, drag_state):
    view = gesture.get_widget()
    selection = view.selection
    event = gesture.get_last_event(None)
    modifiers = event.get_state()[1]
    item, handle = find_item_and_handle_at_point(view, (start_x, start_y))

    # Deselect all items unless CTRL or SHIFT is pressed
    # or the item is already selected.
    if not (
        modifiers & (Gdk.ModifierType.CONTROL_MASK | Gdk.ModifierType.SHIFT_MASK)
        or item in selection.selected_items
    ):
        selection.unselect_all()

    if not item:
        gesture.set_state(Gtk.EventSequenceState.DENIED)
        return

    if (
        not handle
        and item in selection.selected_items
        and modifiers & Gdk.ModifierType.CONTROL_MASK
    ):
        selection.unselect_item(item)
        gesture.set_state(Gtk.EventSequenceState.DENIED)
        return

    selection.focused_item = item
    gesture.set_state(Gtk.EventSequenceState.CLAIMED)

    if handle:
        drag_state.moving = {HandleMove(item, handle, view)}
    else:
        drag_state.moving = set(moving_items(view))

    for moving in drag_state.moving:
        moving.start_move((start_x, start_y))


def find_item_and_handle_at_point(
    view: GtkView, pos: Pos
) -> Tuple[Optional[Item], Optional[Handle]]:
    item, handle = handle_at_point(view, pos)
    return item or item_at_point(view, pos), handle


def moving_items(view):
    """Filter the items that should eventually be moved.

    Returns Move aspects for the items.
    """
    selected_items = set(view.selection.selected_items)
    for item in selected_items:
        # Do not move subitems of selected items
        if not set(ancestors(view.model, item)).intersection(selected_items):
            yield Move(item, view)


def on_drag_update(gesture, offset_x, offset_y, drag_state):
    _, x, y = gesture.get_start_point()
    for moving in drag_state.moving:
        moving.move((x + offset_x, y + offset_y))


def on_drag_end(gesture, offset_x, offset_y, drag_state):
    _, x, y = gesture.get_start_point()
    for moving in drag_state.moving:
        moving.stop_move((x + offset_x, y + offset_y))
    drag_state.moving = set()


def handle_at_point(
    view: GtkView, pos: Pos, distance: int = 6
) -> Union[Tuple[Item, Handle], Tuple[None, None]]:
    """Look for a handle at ``pos`` and return the tuple (item, handle)."""

    def find(item):
        """Find item's handle at pos."""
        v2i = view.get_matrix_v2i(item)
        d = distance_point_point_fast(v2i.transform_distance(0, distance))
        x, y = v2i.transform_point(*pos)

        for h in item.handles():
            if not h.movable:
                continue
            hx, hy = h.pos
            if -d < (hx - x) < d and -d < (hy - y) < d:
                return h

    selection = view.selection

    # The focused item is the preferred item for handle grabbing
    if selection.focused_item:
        h = find(selection.focused_item)
        if h:
            return selection.focused_item, h

    # then try hovered item
    if selection.hovered_item:
        h = find(selection.hovered_item)
        if h:
            return selection.hovered_item, h

    # Last try all items, checking the bounding box first
    x, y = pos
    items = reversed(
        list(
            view.get_items_in_rectangle(
                (x - distance, y - distance, distance * 2, distance * 2)
            )
        )
    )

    for item in items:
        h = find(item)
        if h:
            return item, h
    return None, None



from __future__ import annotations

from typing import Callable, Set, SupportsFloat, Tuple, Union

from gaphas.matrix import Matrix
from gaphas.solver import NORMAL, BaseConstraint, Variable
from gaphas.types import Pos, SupportsFloatPos, TypedProperty


class Position:
    """A point constructed of two `Variable`'s.

    >>> vp = Position(3, 5)
    >>> vp.x, vp.y
    (Variable(3, 20), Variable(5, 20))
    >>> vp.pos
    (Variable(3, 20), Variable(5, 20))
    >>> vp[0], vp[1]
    (Variable(3, 20), Variable(5, 20))
    """

    def __init__(self, x, y, strength=NORMAL):
        self._x = Variable(x, strength)
        self._y = Variable(y, strength)
        self._handlers: Set[Callable[[Position, Pos], None]] = set()
        self._setting_pos = 0

    def add_handler(self, handler: Callable[[Position, Pos], None]) -> None:
        if not self._handlers:
            self._x.add_handler(self._propagate_x)
            self._y.add_handler(self._propagate_y)
        self._handlers.add(handler)

    def remove_handler(self, handler: Callable[[Position, Pos], None]) -> None:
        self._handlers.discard(handler)
        if not self._handlers:
            self._x.remove_handler(self._propagate_x)
            self._y.remove_handler(self._propagate_y)

    def notify(self, oldpos: Pos) -> None:
        for handler in self._handlers:
            handler(self, oldpos)

    def _propagate_x(self, variable, oldval):
        if not self._setting_pos:
            self.notify((oldval, self._y.value))

    def _propagate_y(self, variable, oldval):
        if not self._setting_pos:
            self.notify((self._x.value, oldval))

    @property
    def strength(self) -> int:
        """Strength."""
        return self._x.strength

    def _set_x(self, v: SupportsFloat) -> None:
        self._x.value = v

    x: TypedProperty[Variable, SupportsFloat]
    x = property(lambda s: s._x, _set_x, doc="Position.x")

    def _set_y(self, v: SupportsFloat) -> None:
        self._y.value = v

    y: TypedProperty[Variable, SupportsFloat]
    y = property(lambda s: s._y, _set_y, doc="Position.y")

    def _set_pos(self, pos: Union[Position, SupportsFloatPos]) -> None:
        """Set handle position (Item coordinates)."""
        oldpos = (self._x.value, self._y.value)
        self._setting_pos += 1
        try:
            self._x.value, self._y.value = pos
        finally:
            self._setting_pos -= 1
        self.notify(oldpos)

    pos: TypedProperty[Tuple[Variable, Variable], Union[Position, SupportsFloatPos]]
    pos = property(lambda s: (s._x, s._y), _set_pos, doc="The position.")

    def tuple(self) -> Tuple[float, float]:
        return (self._x.value, self._y.value)

    def __str__(self):
        return f"<{self.__class__.__name__} object on ({self._x}, {self._y})>"

    __repr__ = __str__

    def __getitem__(self, index):
        """Shorthand for returning the x(0) or y(1) component of the point.

        >>> h = Position(3, 5)
        >>> h[0]
        Variable(3, 20)
        >>> h[1]
        Variable(5, 20)
        """
        return (self._x, self._y)[index]

    def __iter__(self):
        return iter((self._x, self._y))

    def __eq__(self, other):
        return isinstance(other, Position) and self.x == other.x and self.y == other.y

    def __lt__(self, other):
        return self[0] < other[0] and self[1] < other[1]


class MatrixProjection(BaseConstraint):
    def __init__(self, pos: Position, matrix: Matrix):
        proj_pos = Position(0, 0, pos.strength)
        super().__init__(proj_pos.x, proj_pos.y, pos.x, pos.y)

        self._orig_pos = pos
        self._proj_pos = proj_pos
        self.matrix = matrix
        self.solve_for(self._proj_pos.x)

    def add_handler(self, handler):
        """Add a callback handler."""
        if not self._handlers:
            self.matrix.add_handler(self._on_matrix_changed)
        super().add_handler(handler)

    def remove_handler(self, handler):
        """Remove a previously assigned handler."""
        super().remove_handler(handler)
        if not self._handlers:
            self.matrix.remove_handler(self._on_matrix_changed)

    @property
    def pos(self) -> Position:
        """The projected position."""
        return self._proj_pos

    def _set_x(self, x):
        self._proj_pos.x = x

    x: TypedProperty[Variable, SupportsFloat]
    x = property(
        lambda s: s._proj_pos.x, _set_x, doc="The projected position's ``x`` part."
    )

    def _set_y(self, y):
        self._proj_pos.y = y

    y: TypedProperty[Variable, SupportsFloat]
    y = property(
        lambda s: s._proj_pos.y, _set_y, doc="The projected position's ``y`` part."
    )

    def mark_dirty(self, var):
        if var is self._orig_pos.x or var is self._orig_pos.y:
            super().mark_dirty(self._orig_pos.x)
            super().mark_dirty(self._orig_pos.y)
        else:
            super().mark_dirty(self._proj_pos.x)
            super().mark_dirty(self._proj_pos.y)

    def solve_for(self, var):
        if var is self._orig_pos.x or var is self._orig_pos.y:
            self._orig_pos.x, self._orig_pos.y = self.matrix.inverse().transform_point(
                *self._proj_pos
            )
        else:
            self._proj_pos.x, self._proj_pos.y = self.matrix.transform_point(
                *self._orig_pos
            )

    def _on_matrix_changed(self, matrix, _orig):
        self.mark_dirty(self._orig_pos.x)
        self.notify()

    def __getitem__(self, index):
        return self._proj_pos[index]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gaphas',
 'gaphas.aspect',
 'gaphas.painter',
 'gaphas.solver',
 'gaphas.tool',
 'gaphas.view']

package_data = \
{'': ['*']}

install_requires = \
['PyGObject>=3.20.0,<4.0.0',
 'pycairo>=1.13.0,<2.0.0',
 'typing-extensions>=3.7.4,<4.0.0']

setup_kwargs = {
    'name': 'gaphas',
    'version': '3.0.0',
    'description': 'Gaphas is a GTK+ based diagramming widget',
    'long_description': '# Gaphas\n[![Build state](https://github.com/gaphor/gaphas/workflows/build/badge.svg)](https://github.com/gaphor/gaphas/actions)\n[![Maintainability](https://api.codeclimate.com/v1/badges/e9837cc647b72119fd11/maintainability)](https://codeclimate.com/github/gaphor/gaphas/maintainability)\n[![Test Coverage](https://api.codeclimate.com/v1/badges/e9837cc647b72119fd11/test_coverage)](https://codeclimate.com/github/gaphor/gaphas/test_coverage)\n![Docs build state](https://readthedocs.org/projects/gaphas/badge/?version=latest)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![standard-readme compliant](https://img.shields.io/badge/readme%20style-standard-brightgreen.svg?style=flat)](https://github.com/RichardLitt/standard-readme)\n[![Gitter](https://img.shields.io/gitter/room/nwjs/nw.js.svg)](https://gitter.im/Gaphor/Lobby?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge&utm_content=badge)\n[![All Contributors](https://img.shields.io/badge/all_contributors-9-orange.svg?style=flat-square)](#contributors)\n\n> Gaphas is the diagramming widget library for Python.\n\n![Gaphas Demo](docs/images/gaphas-demo.gif)\n\nGaphas is a library that provides the user interface component (widget) for drawing diagrams. Diagrams can be drawn to screen and then easily exported to a variety of formats, including SVG and PDF. Want to build an app with chart-like diagrams? Then Gaphas is for you! Use this library to build a tree, network, flowchart, or other diagrams.\n\nThis library is currently being used by [Gaphor](https://github.com/gaphor/gaphor) for UML drawing,\n[RAFCON](https://github.com/DLR-RM/RAFCON) for state-machine based robot control, and [ASCEND](http://ascend4.org/) for solving mathematical models.\n\n## :bookmark_tabs: Table of Contents\n\n- [Background](#background)\n- [Install](#install)\n- [Usage](#usage)\n- [Contributing](#contributing)\n- [License](#license)\n\n## :scroll: Background\n\nGaphas was built to provide the foundational diagramming portions of [Gaphor](https://github.com/gaphor/gaphor). Since Gaphor is built on GTK and Cairo, [PyGObject](https://pygobject.readthedocs.io/) provides access to the GUI toolkit and [PyCairo](https://pycairo.readthedocs.io/) to the 2D graphics library. However, there wasn\'t a project that abstracted these technologies to easily create a diagramming tool. Hence, Gaphas was created as a library to allow others to create a diagramming tool using GTK and Cairo.\n\nHere is how it works:\n\n- Items (Canvas items) can be added to a Canvas.\n- The Canvas maintains the tree structure (parent-child relationships between items).\n- A constraint solver is used to maintain item constraints and inter-item constraints.\n- The item (and user) should not be bothered with things like bounding-box calculations.\n- Very modular--e.g., handle support could be swapped in and swapped out.\n- Rendering using Cairo.\n\nThe main portions of the library include:\n\n- canvas - The main canvas class (container for Items).\n- items - Objects placed on a Canvas.\n- solver - A constraint solver to define the layout and connection of items.\n- gtkview - A view to be used in GTK applications that interacts with users with tools.\n- painters - The workers used to paint items.\n- tools - Tools are used to handle user events (such as mouse movement and button presses).\n- aspects - Provides an intermediate step between tools and items.\n\nGaphas contains default implementations for `Canvas` and `Item`s. There are protocols in place\nto allow you to make your own canvas.\n\n## :floppy_disk: Install\n\nTo install Gaphas, simply use pip:\n\n```bash\n$ pip install gaphas\n```\n\nUse of a\n[virtual environment](https://virtualenv.pypa.io/en/latest/) is highly recommended.\n\n### Development\n\nTo setup a development environment with Linux:\n\n```bash\n$ sudo apt-get install -y python3-dev python3-gi python3-gi-cairo\n    gir1.2-gtk-3.0 libgirepository1.0-dev libcairo2-dev\n$ pip install poetry\n$ poetry install\n```\n\n## :flashlight: Usage\n\nAPI docs and tutorials can be found on [Read the Docs](https://gaphas.readthedocs.io).\n\n## :heart: Contributing\n\nThanks goes to these wonderful people ([emoji key](https://github.com/kentcdodds/all-contributors#emoji-key)):\n\n<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->\n<!-- prettier-ignore -->\n<table><tr><td align="center"><a href="https://github.com/amolenaar"><img src="https://avatars0.githubusercontent.com/u/96249?v=4" width="100px;" alt="Arjan Molenaar"/><br /><sub><b>Arjan Molenaar</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=amolenaar" title="Code">💻</a> <a href="https://github.com/gaphor/gaphas/issues?q=author%3Aamolenaar" title="Bug reports">🐛</a> <a href="https://github.com/gaphor/gaphas/commits?author=amolenaar" title="Documentation">📖</a> <a href="#review-amolenaar" title="Reviewed Pull Requests">👀</a> <a href="#question-amolenaar" title="Answering Questions">💬</a> <a href="#plugin-amolenaar" title="Plugin/utility libraries">🔌</a></td><td align="center"><a href="https://ghuser.io/danyeaw"><img src="https://avatars1.githubusercontent.com/u/10014976?v=4" width="100px;" alt="Dan Yeaw"/><br /><sub><b>Dan Yeaw</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=danyeaw" title="Code">💻</a> <a href="https://github.com/gaphor/gaphas/commits?author=danyeaw" title="Tests">⚠️</a> <a href="#review-danyeaw" title="Reviewed Pull Requests">👀</a> <a href="https://github.com/gaphor/gaphas/issues?q=author%3Adanyeaw" title="Bug reports">🐛</a> <a href="#question-danyeaw" title="Answering Questions">💬</a> <a href="#infra-danyeaw" title="Infrastructure (Hosting, Build-Tools, etc)">🚇</a> <a href="https://github.com/gaphor/gaphas/commits?author=danyeaw" title="Documentation">📖</a></td><td align="center"><a href="https://github.com/wrobell"><img src="https://avatars2.githubusercontent.com/u/105664?v=4" width="100px;" alt="wrobell"/><br /><sub><b>wrobell</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=wrobell" title="Code">💻</a> <a href="https://github.com/gaphor/gaphas/commits?author=wrobell" title="Tests">⚠️</a> <a href="#review-wrobell" title="Reviewed Pull Requests">👀</a></td><td align="center"><a href="https://github.com/jlstevens"><img src="https://avatars3.githubusercontent.com/u/890576?v=4" width="100px;" alt="Jean-Luc Stevens"/><br /><sub><b>Jean-Luc Stevens</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=jlstevens" title="Code">💻</a> <a href="https://github.com/gaphor/gaphas/issues?q=author%3Ajlstevens" title="Bug reports">🐛</a> <a href="https://github.com/gaphor/gaphas/commits?author=jlstevens" title="Documentation">📖</a></td><td align="center"><a href="http://www.franework.de"><img src="https://avatars1.githubusercontent.com/u/1144966?v=4" width="100px;" alt="Franz Steinmetz"/><br /><sub><b>Franz Steinmetz</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=franzlst" title="Code">💻</a> <a href="https://github.com/gaphor/gaphas/issues?q=author%3Afranzlst" title="Bug reports">🐛</a></td><td align="center"><a href="https://github.com/adrianboguszewski"><img src="https://avatars3.githubusercontent.com/u/4547501?v=4" width="100px;" alt="Adrian Boguszewski"/><br /><sub><b>Adrian Boguszewski</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=adrianboguszewski" title="Code">💻</a></td><td align="center"><a href="https://github.com/Rbelder"><img src="https://avatars3.githubusercontent.com/u/15119522?v=4" width="100px;" alt="Rico Belder"/><br /><sub><b>Rico Belder</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/issues?q=author%3ARbelder" title="Bug reports">🐛</a> <a href="#review-Rbelder" title="Reviewed Pull Requests">👀</a></td></tr><tr><td align="center"><a href="http://www.boduch.ca"><img src="https://avatars2.githubusercontent.com/u/114619?v=4" width="100px;" alt="Adam Boduch"/><br /><sub><b>Adam Boduch</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/issues?q=author%3Aadamboduch" title="Bug reports">🐛</a></td><td align="center"><a href="https://github.com/janettech"><img src="https://avatars3.githubusercontent.com/u/13398384?v=4" width="100px;" alt="Janet Jose"/><br /><sub><b>Janet Jose</b></sub></a><br /><a href="https://github.com/gaphor/gaphas/commits?author=janettech" title="Documentation">📖</a></td></tr></table>\n\n<!-- ALL-CONTRIBUTORS-LIST:END -->\n\nThis project follows the [all-contributors](https://github.com/kentcdodds/all-contributors) specification. Contributions of any kind are welcome!\n\n1. Check for open issues or open a fresh issue to start a discussion around a feature idea or a bug.\n    There is a [first-timers-only](https://github.com/gaphor/gaphas/issues?utf8=%E2%9C%93&q=is%3Aissue+is%3Aopen+label%3Afirst-timers-only) tag for issues that should be ideal for people who are not very familiar with the codebase yet.\n2. Fork [the repository](https://github.com/gaphor/gaphas) on GitHub to    start making your changes to the **master**       branch (or branch off of it).\n3. Write a test which shows that the bug was fixed or that the feature\n   works as expected.\n4. Send a pull request and bug the maintainers until it gets merged and\n   published. :smile:\n\nSee [the contributing file](CONTRIBUTING.md)!\n\n## :copyright: License\n\nCopyright (C) Arjan Molenaar and Dan Yeaw\n\nLicensed under the [Apache License 2.0](LICENSE).\n\nSummary: You can do what you like with Gaphas, as long as you include the required notices. This permissive license contains a patent license from the contributors of the code.\n',
    'author': 'Arjan J. Molenaar',
    'author_email': 'gaphor@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gaphas.readthedocs.io/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
